import unittest
from unittest.mock import Mock, MagicMock, patch
import os

import qmi
from qmi.instruments.jpe import Jpe_CPSC, StatusPositionControl
from qmi.core.exceptions import QMI_InstrumentException

readlines_mock = Mock()


class PopenMock:
    stdin = Mock(write=MagicMock())
    stdout = Mock(readlines=readlines_mock)

    def __init__(self, args="wine", stdin=None, stdout=None, cwd=None, env=None):
        pass

    def wait(self):
        pass


class RunMock:
    stdin = Mock(write=MagicMock())
    stdout = ""
    arg_list = []

    def __init__(self, arg_list, stdin=None, stdout=None, encoding=None, check=False):
        self.arg_list = arg_list

    def wait(self):
        pass


class InstanceCreationTestCase(unittest.TestCase):

    def test_default_directory(self):
        """Test that the behaviour is as expected when using default directory while creating instance."""
        if os.path.isdir(Jpe_CPSC.DEFAULT_CPSC_DIR):
            expected_sn = "@1234-567890"
            expected_dir = Jpe_CPSC.DEFAULT_CPSC_DIR
            jpe = Jpe_CPSC(qmi.core.context.QMI_Context("JPE"), "JPE test", expected_sn[1:])  # strip '@' from SN
            self.assertEqual(jpe._serial_number, expected_sn)
            self.assertEqual(jpe._cpsp_dir, expected_dir)

        else:
            with self.assertRaises(QMI_InstrumentException):
                Jpe_CPSC(qmi.core.context.QMI_Context("JPE"), "JPE test", "1234")

    def test_custom_directory(self):
        """Test that a custom directory, set as local directory, does run without exceptions"""
        expected_sn = "@1234-567890"
        expected_dir = os.path.dirname(__file__)
        expected_cmd = os.path.join(expected_dir, "cacli.exe")
        jpe = Jpe_CPSC(qmi.core.context.QMI_Context("JPE"), "JPE test", expected_sn[1:], expected_dir)
        self.assertEqual(jpe._serial_number, expected_sn)
        self.assertEqual(jpe._cpsp_dir, expected_dir)
        self.assertEqual(jpe._cpsp_cmd, expected_cmd)

    def test_faulty_custom_directory(self):
        """Test that a custom directory that does not exist raises an exception"""
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            Jpe_CPSC(qmi.core.context.QMI_Context("JPE"), "JPE test", "1234", r"Q:\utech\QID\QMI")


@patch("sys.platform", "linux1")
class MethodsLinuxTestCase(unittest.TestCase):

    @patch("sys.platform", "linux1")
    def setUp(self):
        # First mock is for the Linux WinE opener
        readlines_mock.side_effect = [["b".encode(), "l".encode(), "a".encode()]]
        self._popen_patcher = patch("qmi.instruments.jpe.cpsc.Popen", new=PopenMock)
        self._popen_patcher.start()
        self.sn = "@1234-567890"
        path = os.path.dirname(__file__)
        self.cmd = os.path.join(path, "cacli.exe")
        self.jpe = Jpe_CPSC(qmi.core.context.QMI_Context("JPE"), "JPE test", self.sn, path)

    @patch("sys.platform", "linux1")
    def tearDown(self):
        if self.jpe._is_open:
            self.jpe.close()

        self._popen_patcher.stop()

    def test_01_open(self):
        """Test opening the instrument in Linux."""
        self.jpe.open()
        self.assertTrue(self.jpe._is_open)

    @patch("sys.platform", "linux1")
    def test_02_close(self):
        """Test closing the instrument in Linux."""
        self.jpe.open()
        self.jpe.close()
        self.assertFalse(self.jpe._is_open)

    def test_03_get_idn(self):
        """Test that the IDN request is correctly made and returns QMI_InstrumentIdentification instance."""
        expected_vendor = "JPE"
        expected_model = "CPSC"
        expected_serial = self.sn
        expected_version = "bla"

        self.jpe.open()
        val = self.jpe.get_idn()
        self.assertEqual(type(val), qmi.core.instrument.QMI_InstrumentIdentification)
        self.assertEqual(val.vendor, expected_vendor)
        self.assertEqual(val.model, expected_model)
        self.assertEqual(val.serial, expected_serial)
        self.assertEqual(val.version, expected_version)
        self.assertEqual(self.jpe._version_string, expected_version)

    def test_04_move(self):
        """See that the move command works as expected on Linux."""
        expected_call = bytes(f"{self.cmd} {self.sn} ".encode()) + b"MOV 2 0 10 100 0 295 CBS10-RLS 1.0\r\n"
        response = "Actuating the stage."
        # self._popen_patcher.stop()
        readlines_mock.side_effect = [[c.encode() for c in response]]
        # self._popen_patcher = patch("qmi.instruments.jpe.cpsc.Popen", new=PopenMock)
        # self._popen_patcher.start()
        self.jpe.open()
        self.jpe.move(2, 0)  # Utilize the default values of the rest of the arguments
        self.jpe._proc_popen.stdin.write.assert_called_with(expected_call)


@patch("sys.platform", "win32")
class MethodsWindowsOnlyTestCase(unittest.TestCase):

    @patch("sys.platform", "win32")
    def setUp(self):
        self._run_patcher = patch("qmi.instruments.jpe.cpsc.run", new=RunMock)
        self._run_patcher.start()
        self.sn = "@1234-567890"
        path = os.path.dirname(__file__)
        self.cmd = os.path.join(path, "cacli.exe")
        self.jpe = Jpe_CPSC(qmi.core.context.QMI_Context("JPE"), "JPE test", self.sn, path)

    @patch("sys.platform", "win32")
    def tearDown(self):
        if self.jpe._is_open:
            self.jpe.close()

        self._run_patcher.stop()

    def test_01_open(self):
        """Test opening the instrument in Windows."""
        self.jpe.open()
        self.assertTrue(self.jpe._is_open)

    def test_02_close(self):
        """Test closing the instrument in Windows."""
        self.jpe.open()
        self.jpe.close()
        self.assertFalse(self.jpe._is_open)

    def test_03_get_idn(self):
        """Test that the IDN request is correctly made and returns QMI_InstrumentIdentification instance."""
        expected_vendor = "JPE"
        expected_model = "CPSC"
        expected_serial = self.sn
        expected_version = "vla"
        self._run_patcher.new.stdout = "vla"

        self.jpe.open()
        val = self.jpe.get_idn()
        self.assertEqual(type(val), qmi.core.instrument.QMI_InstrumentIdentification)
        self.assertEqual(val.vendor, expected_vendor)
        self.assertEqual(val.model, expected_model)
        self.assertEqual(val.serial, expected_serial)
        self.assertEqual(val.version, expected_version)
        self.assertEqual(self.jpe._version_string, expected_version)

    def test_04_get_all_modules_info(self):
        """Test getting all module information."""
        expected_module_a = "moduleA"
        expected_module_b = "moduleB"
        self._run_patcher.new.stdout = f"{expected_module_a},{expected_module_b}"
        self.jpe.open()
        val = self.jpe.get_info_all_modules()
        self.assertEqual(val[0], expected_module_a)
        self.assertEqual(val[1], expected_module_b)

    def test_05_get_module_info(self):
        """Test getting specific module info."""
        expected_module_a = "moduleA"
        self._run_patcher.new.stdout = f" {expected_module_a} "  # see that spaces are also stripped
        self.jpe.open()
        val = self.jpe.get_info_module(1)
        self.assertEqual(val, expected_module_a)

    def test_06_request_fail_safe_state(self):
        """See that the requesting the error works and a string is returned."""
        expected_error = "No Error"
        self._run_patcher.new.stdout = f" {expected_error}"  # see that spaces are also stripped
        self.jpe.open()
        val = self.jpe.request_fail_safe_state("1")
        self.assertEqual(val, expected_error)

    def test_07_move(self):
        """See that the move command works as expected on Windows."""
        expected_args = [f"{self.cmd}", f"{self.sn}", "MOV", "2", "0", "10", "100", "0", "295", "CBS10-RLS", "1.0"]
        response = "Actuating the stage."
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.move(2, 0)  # Utilize the default values of the rest of the arguments
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_08_stop(self):
        """Test stop command."""
        address = 1
        expected_args = [f"{self.cmd}", f"{self.sn}", "STP", f"{address}"]
        response = "Stopping the stage."
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.stop(address)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_09_enable_scan_mode(self):
        """Test enable scan mode command."""
        address = 1
        value = 1023
        expected_args = [f"{self.cmd}", f"{self.sn}", "SDC", f"{address}", f"{value}"]
        response = "Scan mode enabled."
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.enable_scan_mode(address, value)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_10_use_external_input(self):
        """Test setting external input mode."""
        address = 1
        direction = 0
        expected_args = [f"{self.cmd}", f"{self.sn}", "EXT", f"{address}", f"{direction}", "10", "100", "295",
                         "CBS10-RLS", "1.0"]
        response = "External mode enabled."
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.use_external_input(address, direction)  # Utilize the default values of the rest of the arguments
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_11_get_current_position(self):
        """Test that current position value is returned as a float."""
        address = 3
        channel = 2
        expected_response = 1.0
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        pos = self.jpe.get_current_position(address, channel)
        self.assertEqual(pos, expected_response)

    def test_12_get_current_position_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        channel = 2
        unexpected_response = "ERROR: SomeError"
        expected_error_msg = f"Erroneous response from position query SDC {address} {channel} CBS10-RLS:" +\
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.get_current_position(address, channel)
            self.assertEqual(exc[2], expected_error_msg)

    def test_13_get_current_position_of_all_3_channels(self):
        """Test that current position values are returned as a list of floats."""
        address = 3
        expected_response = [1.0, 2.0, 3.0]
        response = f"{expected_response[0]},{expected_response[1]},{expected_response[2]}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        pos = self.jpe.get_current_position_of_all_3_channels(address)
        self.assertListEqual(pos, expected_response)

    def test_14_get_current_position_of_all_3_channels_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        unexpected_response = "1.0,3.0,ERROR: SomeError"
        expected_error_msg = f"Erroneous response from position query PGVA {address} CBS10-RLS:" + \
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.get_current_position_of_all_3_channels(address)
            self.assertEqual(exc[2], expected_error_msg)

    def test_15_set_negative_end_stop(self):
        """Test setting negative end stop command."""
        address = 3
        channel = 2
        expected_args = [f"{self.cmd}", f"{self.sn}", "MIS", f"{address}", f"{channel}"]
        expected_response = "Minimum position set."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.set_negative_end_stop(address, channel)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_16_set_positive_end_stop(self):
        """Test setting positive end stop command."""
        address = 3
        channel = 2
        expected_args = [f"{self.cmd}", f"{self.sn}", "MAS", f"{address}", f"{channel}"]
        expected_response = "Maximum position set."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.set_positive_end_stop(address, channel)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_17_read_negative_end_stop(self):
        """Test reading negative end stop value command returns float."""
        address = 3
        channel = 2
        expected_response = -1.0E-7
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        pos = self.jpe.read_negative_end_stop(address, channel)
        self.assertEqual(pos, expected_response)

    def test_18_read_negative_end_stop_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        channel = 2
        unexpected_response = "Minimum and maximum end-stops reset."
        expected_error_msg = f"Erroneous response from position query MIR {address} {channel} CBS10-RLS:" + \
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.read_negative_end_stop(address, channel)
            self.assertEqual(exc[2], expected_error_msg)

    def test_19_read_positive_end_stop(self):
        """Test reading positive end stop value command returns float."""
        address = 3
        channel = 2
        expected_response = 1.0E-7
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        pos = self.jpe.read_positive_end_stop(address, channel)
        self.assertEqual(pos, expected_response)

    def test_20_read_positive_end_stop_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        channel = 2
        unexpected_response = "Minimum and maximum end-stops reset."
        expected_error_msg = f"Erroneous response from position query MAR {address} {channel} CBS10-RLS:" + \
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.read_positive_end_stop(address, channel)
            self.assertEqual(exc[2], expected_error_msg)

    def test_21_reset_end_stops(self):
        """Test resetting the end stop values command."""
        address = 3
        channel = 2
        expected_args = [f"{self.cmd}", f"{self.sn}", "MMR", f"{address}", f"{channel}"]
        expected_response = "Minimum and maximum end-stops reset."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.reset_end_stops(address, channel)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_22_set_excitation_duty_cycle(self):
        """Test setting excitation duty cycle command."""
        address = 3
        duty = 30
        expected_args = [f"{self.cmd}", f"{self.sn}", "EXS", f"{address}", f"{duty}"]
        expected_response = "Excitation duty cycle set."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.set_excitation_duty_cycle(address, duty)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_23_read_excitation_duty_cycle(self):
        """Test reading excitation duty cycle command returns integer."""
        address = 3
        expected_response = 30
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        pos = self.jpe.read_excitation_duty_cycle(address)
        self.assertEqual(pos, expected_response)

    def test_24_read_excitation_duty_cycle_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        unexpected_response = "30.0"
        expected_error_msg = f"Erroneous response from position query EXR {address}:" + \
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.read_excitation_duty_cycle(address)
            self.assertEqual(exc[2], expected_error_msg)

    def test_25_save_rsm_setting(self):
        """Test save RSM setting command."""
        address = 3
        expected_args = [f"{self.cmd}", f"{self.sn}", "RSS", f"{address}"]
        expected_response = "Settings stored in flash."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.save_rsm_settings(address)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_26_get_current_counter_value(self):
        """Test get current counter value command returns an integer."""
        address = 3
        channel = 2
        expected_response = 12345
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        pos = self.jpe.get_current_counter_value(address, channel)
        self.assertEqual(pos, expected_response)

    def test_27_get_current_counter_value_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        channel = 2
        unexpected_response = "1234.5"
        expected_error_msg = f"Erroneous response from position query CGV {address} {channel}:" + \
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.get_current_counter_value(address, channel)
            self.assertEqual(exc[2], expected_error_msg)

    def test_28_reset_counter_value(self):
        """Test reset counter value command."""
        address = 3
        channel = 2
        expected_args = [f"{self.cmd}", f"{self.sn}", "CSZ", f"{address}", f"{channel}"]
        expected_response = "Position counter set to 0."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.reset_counter_value(address, channel)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_29_get_current_encoder_signal_value(self):
        """Test get current encoder signal value command returns an integer"""
        address = 3
        channel = 2
        expected_response = 123
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        pos = self.jpe.get_current_encoder_signal_value(address, channel)
        self.assertEqual(pos, expected_response)

    def test_30_get_current_encoder_signal_value_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        channel = 2
        unexpected_response = "1.23"
        expected_error_msg = f"Erroneous response from position query DGV {address} {channel}:" + \
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.get_current_encoder_signal_value(address, channel)
            self.assertEqual(exc[2], expected_error_msg)

    def test_31_auto_oem_calibration(self):
        """Test auto OEM calibration command."""
        address = 1
        channel = 0
        cadm2 = 6
        temp = 298
        expected_args = [f"{self.cmd}", f"{self.sn}", "OEMC", f"{address}", f"{channel}", f"{cadm2}", f"{temp}",
                         "CBS10-RLS"]
        response = "Channel calibrated."
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.auto_oem_calibration(address,channel, cadm2, temp)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_32_request_calibration_values(self):
        """Test requesting calibration values command returns a list of integers."""
        address = 3
        channel = 2
        expected_response = [1, 12, 123]
        response = f"{expected_response[0]},{expected_response[1]},{expected_response[2]}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        values = self.jpe.request_calibration_values(address, channel)
        self.assertListEqual(values, expected_response)

    def test_33_request_calibration_values_excepts(self):
        """See that an exception is raised with an erroneous response."""
        address = 3
        channel = 2
        unexpected_response = "1,12.0,123"
        expected_error_msg = f"Erroneous response from calibration value query MLS {address} {channel}:" + \
                             f"{unexpected_response}"
        response = f"{unexpected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.jpe.request_calibration_values(address, channel)
            self.assertEqual(exc[2], expected_error_msg)

    def test_34_set_detector_gain(self):
        """Test set detector gain command."""
        address = 3
        channel = 2
        gain = 30
        expected_args = [f"{self.cmd}", f"{self.sn}", "DSG", f"{address}", f"{channel}", f"{gain}"]
        expected_response = "Detector gain set."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.set_detector_gain(address, channel, gain)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_35_set_upper_threshold(self):
        """Test setting upper threshold command."""
        address = 3
        channel = 2
        ut = 130
        expected_args = [f"{self.cmd}", f"{self.sn}", "DSH", f"{address}", f"{channel}", f"{ut}"]
        expected_response = "Detector upper threshold set."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.set_upper_threshold(address, channel, ut)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_36_set_lower_threshold(self):
        """Test setting lower threshold."""
        address = 3
        channel = 2
        lt = 13
        expected_args = [f"{self.cmd}", f"{self.sn}", "DSL", f"{address}", f"{channel}", f"{lt}"]
        expected_response = "Detector lower threshold set."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.set_lower_threshold(address, channel, lt)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_37_save_detector_settings(self):
        """Test saving detector settings command."""
        address = 2
        expected_args = [f"{self.cmd}", f"{self.sn}", "MSS", f"{address}"]
        expected_response = "Detector settings stored in flash."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.save_detector_settings(address)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_38_enable_servodrive(self):
        """Test enabling servo drive command."""
        freq = 10
        stage = "Stage"
        df = 2.0
        temp = 4
        expected_args = [f"{self.cmd}", f"{self.sn}", "FBEN", f"{stage}", f"{freq}", f"{stage}", f"{freq}", f"{stage}",
                         f"{freq}", f"{df}", f"{temp}"]
        expected_response = "Control loop enabled."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.enable_servodrive(stage, freq, stage, freq, stage, freq, df, temp)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_39_disable_servodrive(self):
        """Test disabling servo drive command."""
        expected_args = [f"{self.cmd}", f"{self.sn}", "FBXT"]
        expected_response = "Control loop disabled."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.disable_servodrive()
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_40_move_to_setpoint(self):
        """Test move to setpoint command."""
        sp1 = 1E-9
        sp2 = 2E-9
        sp3 = 3E-9
        abs = 0
        expected_args = [f"{self.cmd}", f"{self.sn}", "FBCS", f"{sp1}", f"{abs}", f"{sp2}", f"{abs}", f"{sp3}",
                         f"{abs}"]
        expected_response = "Control loop setpoints set."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.move_to_setpoint(sp1, sp2, sp3, abs)
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_41_emergency_stop(self):
        """Test emergency stop command."""
        expected_args = [f"{self.cmd}", f"{self.sn}", "FBES"]
        expected_response = "Control loop emergency stop enabled."
        response = f"{expected_response}"
        self._run_patcher.new.stdout = response
        self.jpe.open()
        self.jpe.emergency_stop()
        self.assertEqual(self.jpe._proc_run.arg_list, expected_args)

    def test_42_get_status_position_control(self):
        """Test get status position control returns a string of comma-separated statuses"""
        response = "1,0,0,0,0,-8528,-11864,42770"
        self._run_patcher.new.stdout = response
        expected_status = StatusPositionControl(*map(int, response.split(",")))
        self.jpe.open()
        status = self.jpe.get_status_position_control()
        self.assertEqual(status, expected_status)


if __name__ == '__main__':
    unittest.main()



"""Unit tests for a Newport single axis motion controller."""

from typing import cast
import unittest
from unittest.mock import Mock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.utils.context_managers import start_stop
from qmi.instruments.newport import Newport_Smc100Pp, Newport_Smc100Cc, Newport_ConexCc
from qmi.instruments.newport.actuators import TRA12CC, TRB6CC, CMA25CCL, UTS100PP
from qmi.instruments.newport.single_axis_motion_controller import Newport_SingleAxisMotionController


class TestDerivingClassCase(unittest.TestCase):
    """Test that the creation of child classes of Newport_Single_Axis_Motion_Controller work as expected."""

    def setUp(self):
        # Add patches
        patcher = patch(
            "qmi.instruments.newport.single_axis_motion_controller.create_transport", spec=QMI_TcpTransport
        )
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = patch("qmi.instruments.newport.single_axis_motion_controller.ScpiProtocol", autospec=True)
        self._scpi_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)

    def test_smc100pp(self):
        """
        Test SMC100PP
        """
        expected_rpc_class = "qmi.instruments.newport.smc_100pp.Newport_SMC100PP"
        with start_stop(qmi, "TestSMC100PP", console_loglevel="CRITICAL"):
            # Make DUT
            self.instr: Newport_Smc100Pp = qmi.make_instrument(
                "pp_controller", Newport_Smc100Pp, "Beverly_Hills", "FT5TMFGL", {1: UTS100PP}, 90210
            )
            # Test __init__ is of correct module
            self.assertIn(expected_rpc_class, str(self.instr.__init__))

    def test_smc100cc(self):
        """
        Test SMC100CC
        """
        expected_rpc_class = "qmi.instruments.newport.smc_100cc.Newport_SMC100CC"
        with start_stop(qmi, "TestSMC100CC", console_loglevel="CRITICAL"):
            # Make DUT
            self.instr: Newport_Smc100Cc = qmi.make_instrument(
                "sam_controller",
                Newport_Smc100Cc,
                "Beverly_Hills",
                "FT5TMFGL",
                {1: TRA12CC, 2: TRB6CC, 3: CMA25CCL},
                90210,
            )
            # Test __init__ is of correct module
            self.assertIn(expected_rpc_class, str(self.instr.__init__))

    def test_conex_cc(self):
        """
        Test Conex CC
        """
        expected_rpc_class = "qmi.instruments.newport.conex_cc.Newport_ConexCC"
        with start_stop(qmi, "TestConexCC", console_loglevel="CRITICAL"):
            # Make DUT
            self.instr: Newport_ConexCc = qmi.make_instrument(
                "sam_controller",
                Newport_ConexCc,
                "Beverly_Hills",
                "FT5TMFGL",
                {1: TRA12CC, 2: TRB6CC, 3: CMA25CCL},
                90210,
            )
            # Test __init__ is of correct module
            self.assertIn(expected_rpc_class, str(self.instr.__init__))


class TestSingleAxisMotionController(unittest.TestCase):
    """
    Tests for the single axis motion controller.
    """

    TRANSPORT_STR = "/dev/cu.usbserial-FT5TMFGL"

    def setUp(self):
        qmi.start("TestSamControllerContext", console_loglevel="CRITICAL")
        # Add patches
        patcher = patch(
            "qmi.instruments.newport.single_axis_motion_controller.create_transport", spec=QMI_TcpTransport
        )
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch("qmi.instruments.newport.single_axis_motion_controller.ScpiProtocol", autospec=True)
        self._scpi_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUTs
        self.controller_address = 1
        Newport_SingleAxisMotionController.PW0_EXEC_TIME = 0.1
        Newport_SingleAxisMotionController.COMMAND_EXEC_TIME = 0.01
        self.instr: Newport_SingleAxisMotionController = qmi.make_instrument(
            "sam_controller",
            Newport_SingleAxisMotionController,
            self.TRANSPORT_STR,
            "FT5TMFGL",
            {1: TRA12CC, 2: TRB6CC, 3: CMA25CCL},
            90210,
        )
        self.instr = cast(Newport_SingleAxisMotionController, self.instr)
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_reset_without_controller_address_resets(self):
        """Test reset."""
        self._scpi_mock.ask.return_value = "@"
        self.instr.reset()

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}RS\r\n")
        self._scpi_mock.ask.assert_called_once_with(f"{self.controller_address}TE\r\n")

    def test_reset_with_controller_address_resets(self):
        """Test reset with controller address."""
        self._scpi_mock.ask.return_value = "@"
        self.instr.reset(3)

        self._scpi_mock.write.assert_called_once_with("3RS\r\n")
        self._scpi_mock.ask.assert_called_once_with("3TE\r\n")

    def test_build_command_with_invalid_controller_address_throws_exception(self):
        """Test build command with invalid controller address."""
        with self.assertRaises(QMI_InstrumentException):
            self.instr.reset(4)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_idn_address_gets_idn(self):
        """Test get idn."""
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}VE CONEX-CC 2.0.1", "@"]
        expected_calls = [call(f"{self.controller_address}VE\r\n"), call(f"{self.controller_address}TE\r\n")]
        info = self.instr.get_idn()
        self.assertEqual(info.vendor, "Newport")
        self.assertEqual(info.version, "2.0.1")
        self.assertEqual(info.serial, "FT5TMFGL")
        self.assertEqual(info.model, "CONEX-CC")

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_get_stage_identifier(self):
        """Test get stage identifier."""
        state = "14"  # Only need to exit from CONFIGURATION mode.
        expected_identifier = "URS100CC"
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{expected_identifier}", "@"]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}ID?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        idn = self.instr.get_stage_identifier()
        self.assertEqual(idn, expected_identifier)

        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)

    def test_get_positioner_error_and_state_gets_positioner_error_and_state(self):
        """Test positioner error and state."""
        self._scpi_mock.ask.return_value = f"{self.controller_address}TS00001E"
        err_and_state = self.instr.get_positioner_error_and_state()

        self.assertEqual(err_and_state[1], "HOMING.")
        self.assertEqual(len(err_and_state[0]), 0)
        self._scpi_mock.ask.assert_called_once_with(f"{self.controller_address}TS\r\n")

    def test_get_positioner_error_and_state_with_an_error_bit(self):
        """Test positioner error gets added to 'errors' and returned, with state."""
        expected_errors = [
            Newport_SingleAxisMotionController.POSITIONER_ERROR_TABLE[9],
            Newport_SingleAxisMotionController.POSITIONER_ERROR_TABLE[12],
            Newport_SingleAxisMotionController.POSITIONER_ERROR_TABLE[13],
        ]
        self._scpi_mock.ask.return_value = f"{self.controller_address}TS004C1E"
        errors, state = self.instr.get_positioner_error_and_state()

        self.assertEqual(state, "HOMING.")
        self.assertListEqual(expected_errors, errors)
        self._scpi_mock.ask.assert_called_once_with(f"{self.controller_address}TS\r\n")

    def test_home_search_homes(self):
        """Test home search."""
        self._scpi_mock.ask.return_value = "@"
        self.instr.home_search()

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}OR\r\n")
        self._scpi_mock.ask.assert_called_once_with(f"{self.controller_address}TE\r\n")

    def test_set_home_search_timeout_sets_timeout(self):
        """Test set home search timeout."""
        state = "14"  # Only need to exit from CONFIGURATION mode.
        timeout = 13
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}OT%s\r\n" % timeout),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        self.instr.set_home_search_timeout(timeout)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_home_search_timeout_with_default_timeout_sets_timeout(self):
        """Test set home search timeout with controller address."""
        timeout = Newport_SingleAxisMotionController.DEFAULT_HOME_SEARCH_TIMEOUT
        state = "14"  # Only need to exit from CONFIGURATION mode.
        self._scpi_mock.ask.side_effect = [f"3TS0000{state}", "@", "@"]
        expected_write_calls = [call(f"3OT{timeout}\r\n"), call("3PW0\r\n")]
        expected_ask_calls = [call("3TS\r\n"), call("3TE\r\n")]
        self.instr.set_home_search_timeout(controller_address=3)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_home_search_timeout_gets_timeout(self):
        """Test get home search timeout."""
        expected_timeout = 10
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}OT%s" % expected_timeout, "@"]
        expected_calls = [call(f"{self.controller_address}OT?\r\n"), call(f"{self.controller_address}TE\r\n")]
        actual_timeout = self.instr.get_home_search_timeout()

        self.assertEqual(expected_timeout, actual_timeout)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_get_home_search_timeout_with_error_throws_exception(self):
        """Test get home search timeout with error."""
        expected_timeout = 10
        expected_error = ["Error", "C", ": Command not allowed"]
        self._scpi_mock.ask.side_effect = [
            f"{self.controller_address}OT%s" % expected_timeout,
            expected_error[1],
            f"{self.controller_address}TBD{expected_error[2]}",
        ]
        expected_calls = [
            call(f"{self.controller_address}OT?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TBC\r\n"),
        ]
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.instr.get_home_search_timeout()

        self._scpi_mock.ask.assert_has_calls(expected_calls)
        self.assertEqual(str(exc.exception), expected_error[0] + f" {expected_error[1]}" + expected_error[2])

    def test_move_absolute_more_than_max_thrown_exception(self):
        """Test move absolute more than max."""
        self._scpi_mock.ask.return_value = "@"

        with self.assertRaises(QMI_InstrumentException):
            self.instr.move_absolute(1000000)

        self._scpi_mock.write.assert_not_called()

    def test_move_absolute_less_than_min_throws_exception(self):
        """Test move absolute less than min."""
        self._scpi_mock.ask.return_value = "@"

        with self.assertRaises(QMI_InstrumentException):
            self.instr.move_absolute(-1)

        self._scpi_mock.write.assert_not_called()

    def test_move_absolute_moves(self):
        """Test move absolute."""
        state = "32"  # Moves only in "READY" state
        pos = 1
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        self._scpi_mock.ask.return_value = "@"
        self.instr.move_absolute(pos)

        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)
        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}PA%s\r\n" % pos)

    def test_move_absolute_excepts_as_not_in_ready_state(self):
        """Test move absolute."""
        state = "3C"  # Moves only in "READY" state
        pos = 1
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n")]
        self._scpi_mock.ask.return_value = "@"
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.instr.move_absolute(pos)

        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)
        self._scpi_mock.write.assert_not_called()

    def test_get_setpoint_gets_setpoint(self):
        """Test get setpoint."""
        expected_sp = 1.2345
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}TH %s" % expected_sp, "@"]
        expected_calls = [call(f"{self.controller_address}TH\r\n"), call(f"{self.controller_address}TE\r\n")]
        actual_sp = self.instr.get_setpoint()
        self.assertEqual(actual_sp, expected_sp)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_get_position_gets_position(self):
        """Test get position."""
        expected_pos = 1.2345
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}TP %s" % expected_pos, "@"]
        expected_calls = [call(f"{self.controller_address}TP\r\n"), call(f"{self.controller_address}TE\r\n")]
        actual_pos = self.instr.get_position()
        self.assertEqual(actual_pos, expected_pos)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_move_relative_with_negative_displacement_moves(self):
        """Test move relative with negative displacement and controller address."""
        state = "34"  # READY
        pos = -0.1
        self._scpi_mock.ask.side_effect = [f"3TS0000{state}", "@"]
        self.instr.move_relative(pos, 3)

        self._scpi_mock.write.assert_called_once_with(f"3PR{pos}\r\n")
        self._scpi_mock.ask.assert_any_call("3TE\r\n")

    def test_move_relative_moves(self):
        """Test move relative with controller address."""
        state = "34"  # READY
        pos = 1
        self._scpi_mock.ask.side_effect = [f"3TS0000{state}", "@"]
        self.instr.move_relative(pos, 3)

        self._scpi_mock.write.assert_called_once_with(f"3PR{pos}\r\n")
        self._scpi_mock.ask.assert_any_call("3TE\r\n")

    def test_move_relative_excepts(self):
        """Test move relative excepts if controller not in READY state."""
        state = "3C"  # DISABLE
        pos = 1
        self._scpi_mock.ask.side_effect = [f"2TS0000{state}"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.move_relative(pos, 2)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_any_call("2TS\r\n")

    def test_move_relative_less_than_min_throws_exception(self):
        """Test move relative less than min."""
        with self.assertRaises(QMI_InstrumentException):
            self.instr.move_relative(0.0000000001)

        self._scpi_mock.write.assert_not_called()

    def test_get_motion_time(self):
        """Test get motion time."""
        displacement = 3.21
        expected_time = 1.23
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}PT %s" % expected_time, "@"]
        expected_calls = [
            call(f"{self.controller_address}PT{displacement}\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_time = self.instr.get_motion_time(displacement)
        self.assertEqual(actual_time, expected_time)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_set_configuration_state_sets_configuration_state(self):
        """Test set configuration state."""
        self.instr.set_configuration_state(True)

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}PW1\r\n")
        self._scpi_mock.ask.assert_not_called()

        self._scpi_mock.write.reset_mock()
        self._scpi_mock.ask.reset_mock()

        self.instr.set_configuration_state(False)

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}PW0\r\n")
        self._scpi_mock.ask.assert_not_called()

    def test_get_configuration_state_gets_configuration_state(self):
        """Test get configuration state with controller address."""
        contr_addr = 3
        expected_state = False
        self._scpi_mock.ask.side_effect = [f"{contr_addr}PW{int(expected_state)}", "@"]

        state = self.instr.is_in_configuration_state(controller_address=contr_addr)

        self.assertEqual(state, expected_state)
        self._scpi_mock.ask.assert_any_call(f"{contr_addr}PW?\r\n")
        self._scpi_mock.ask.assert_any_call(f"{contr_addr}TE\r\n")

    def test_set_disable_state_sets_disable_state(self):
        """Test set disable state."""
        self._scpi_mock.ask.return_value = "@"
        # 1. Test setting state to DISABLE
        self.instr.set_disable_state(True)

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}MM1\r\n")
        self._scpi_mock.ask.assert_called_once_with("1TE\r\n")

        # Test setting state to READY
        self._scpi_mock.write.reset_mock()
        self._scpi_mock.ask.reset_mock()
        self.instr.set_disable_state(False)

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}MM0\r\n")
        self._scpi_mock.ask.assert_called_once_with("1TE\r\n")

    def test_get_disable_state_gets_disable_state(self):
        """Test get disable state with controller address."""
        contr_addr = 2
        expected_state = False
        self._scpi_mock.ask.side_effect = [f"{contr_addr}MM{int(expected_state)}", "@"]

        state = self.instr.is_in_disable_state(controller_address=contr_addr)

        self.assertEqual(state, expected_state)
        self._scpi_mock.ask.assert_any_call(f"{contr_addr}MM?\r\n")
        self._scpi_mock.ask.assert_any_call(f"{contr_addr}TE\r\n")

    def test_get_error_with_no_error_gets_error(self):
        """Test get error with no error."""
        expected_error_code = "@"
        expected_error_message = "No error"
        self._scpi_mock.ask.return_value = expected_error_code

        actual_error = self.instr.get_error()

        self.assertEqual(actual_error[0], expected_error_code)
        self.assertEqual(actual_error[1], expected_error_message)

        self._scpi_mock.ask.assert_called_once_with(f"{self.controller_address}TE\r\n")

    def test_get_error_with_error_gets_error(self):
        """Test get error with an error."""
        expected_error_code = "C"
        expected_error_message = "Parameter missing or out of range"
        self._scpi_mock.ask.side_effect = [expected_error_code, "Error: " + expected_error_message]
        expected_calls = [
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TB%s\r\n" % expected_error_code),
        ]
        actual_error = self.instr.get_error()

        self.assertEqual(actual_error[0], expected_error_code)
        self.assertEqual(actual_error[1], expected_error_message)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_get_acceleration_gets_acceleration_in_READY_state(self):
        """Test get acceleration in READY state."""
        state = "32"
        expected_acc = 1.5
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}AC %s" % expected_acc, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}AC?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_acc = self.instr.get_acceleration()
        self.assertEqual(actual_acc, expected_acc)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_acceleration_gets_acceleration_in_CONFIGURATION_state(self):
        """Test get acceleration in CONFIGURATION state."""
        state = "14"  # Only need to exit from CONFIGURATION mode.
        expected_acc = 1.5
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}",
            f"1TS0000{state}",
            f"{self.controller_address}AC %s" % expected_acc,
            "@",
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}AC?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_acc = self.instr.get_acceleration()
        self.assertEqual(actual_acc, expected_acc)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_acceleration_with_persist_sets_acceleration(self):
        """Test set acceleration with persist = True."""
        state = "33"
        expected_acc = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}RS\r\n"),
            call(f"{self.controller_address}PW1\r\n"),
            call(f"{self.controller_address}AC{expected_acc}\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        self.instr.set_acceleration(expected_acc, True)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_acceleration_without_persist_sets_acceleration(self):
        """Test set acceleration without persisting, i.e. without entering the configuration state."""
        state = "33"
        expected_acc = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}AC{expected_acc}\r\n"),
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        self.instr.set_acceleration(expected_acc)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_acceleration_outside_bounds_throws_exception(self):
        """Test set acceleration outisde of allowed range raise exception."""
        self._scpi_mock.ask.return_value = "@"
        out_of_bounds = [1e-6, 1e12]

        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_acceleration(value)

        self._scpi_mock.write.assert_not_called()

    def test_set_velocity_more_than_max_throws_exception(self):
        """Test set velocity more than max."""
        self._scpi_mock.ask.return_value = "@"

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_velocity(1000000)

        self._scpi_mock.write.assert_not_called()

    def test_set_velocity_less_than_min_throws_exception(self):
        """Test set velocity less than min."""
        self._scpi_mock.ask.return_value = "@"

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_velocity(0.0000000001)

        self._scpi_mock.write.assert_not_called()

    def test_get_velocity_gets_velocity(self):
        """Test get velocity."""
        state = "34"  # Set to READY state to take the shortest path
        expected_vel = 1.5
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}VA %s" % expected_vel, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]

        actual_vel = self.instr.get_velocity()

        self.assertEqual(actual_vel, expected_vel)
        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_velocity_gets_velocity_in_configuration(self):
        """Test get velocity in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state.
        expected_vel = 1.5
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}",
            f"1TS0000{state}",
            f"{self.controller_address}VA %s" % expected_vel,
            "@",
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}VA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        actual_vel = self.instr.get_velocity()
        self.assertEqual(actual_vel, expected_vel)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_velocity_with_persist_sets_velocity(self):
        """Test set velocity with persisting the velocity."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        vel = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}VA%s\r\n" % vel),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        self.instr.set_velocity(vel, True)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_velocity_without_persist_sets_velocity(self):
        """Test set velocity without persisting, i.e. without entering the configuration state."""
        state = "33"
        vel = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_write_calls = [
            call(f"{self.controller_address}VA{vel}\r\n"),
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        self.instr.set_velocity(vel)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_jerk_time_gets_jerk_time(self):
        """Test get jerk_time."""
        state = "34"  # READY state
        expected_jerk = 1.5
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}JR %s" % expected_jerk, "@"]
        expected_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}JR?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_jerk = self.instr.get_jerk_time()
        self.assertEqual(actual_jerk, expected_jerk)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_get_jerk_time_gets_jerk_time_in_configuration(self):
        """Test get jerk_time in CONFIGURATION state."""
        state = "14"  # CONFIGURATION state
        expected_jerk = 1.5
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}",
            f"1TS0000{state}",
            f"{self.controller_address}JR %s" % expected_jerk,
            "@",
        ]
        expected_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}JR?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_jerk = self.instr.get_jerk_time()
        self.assertEqual(actual_jerk, expected_jerk)

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}PW0\r\n")
        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_set_jerk_time_with_persist_sets_jerk_time(self):
        """Test set jerk_time with persist = True."""
        state = "14"  # CONFIGURATION state
        jerk = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [
            call(f"{self.controller_address}JR{jerk}\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        self.instr.set_jerk_time(jerk, True)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_jerk_time_without_persist_sets_jerk_time(self):
        """Test set jerk_time without persisting, i.e. without entering the configuration state."""
        state = "34"  # READY state
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        jerk = 0.004
        self._scpi_mock.ask.return_value = "@"
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]

        self.instr.set_jerk_time(jerk)

        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}JR{jerk}\r\n")
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_jerk_time_outside_bounds_throws_exception(self):
        """Test set jerk_time outisde of allowed range raise exception."""
        self._scpi_mock.ask.return_value = "@"
        out_of_bounds = [0.001, 1e12]

        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_jerk_time(value)

        self._scpi_mock.write.assert_not_called()

    def test_stop_motion_stops_motion(self):
        """Test stop motion."""
        self._scpi_mock.ask.return_value = "@"
        self.instr.stop_motion(3)

        self._scpi_mock.write.assert_called_once_with("3ST\r\n")
        self._scpi_mock.ask.assert_called_once_with("3TE\r\n")

    def test_set_backlash_compensation_sets_compensation(self):
        """Test set backlash compensation with controller address."""
        state = "14"  # CONFIGURATION state
        hyst_comp = 0.0
        comp = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"3BH %s" % hyst_comp, "@", "@"]
        expected_ask_calls = [call(f"3TS\r\n"), call(f"3BH?\r\n"), call(f"3TE\r\n"), call(f"3TE\r\n")]
        expected_write_calls = [call(f"3BA{comp}\r\n"), call(f"3PW0\r\n")]
        self.instr.set_backlash_compensation(comp, 3)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_backlash_compensation_excepts_because_hysteresis_compensation(self):
        """Test set backlash compensation exits because hysteresis compensation is enabled."""
        state = "32"  # Will call reset to get into CONFIGURATION state
        hyst_comp = 0.1
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", f"{self.controller_address}BH %s" % hyst_comp, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}BH?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [
            call(f"{self.controller_address}RS\r\n"),
            call(f"{self.controller_address}PW1\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        comp = 0.004
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_backlash_compensation(comp)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_backlash_compensation_gets_compensation(self):
        """Test get backlash compensation."""
        state = "32"  # Will call reset to get into CONFIGURATION state
        expected_compensation = 10
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}",
            "@",
            f"{self.controller_address}BA %s" % expected_compensation,
            "@",
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}BA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [
            call(f"{self.controller_address}RS\r\n"),
            call(f"{self.controller_address}PW1\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        actual_comp = self.instr.get_backlash_compensation()
        self.assertEqual(expected_compensation, actual_comp)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_hysteresis_compensation_sets_compensation(self):
        """Test set hysteresis compensation with controller address."""
        state = "14"  # CONFIGURATION state
        back_comp = 0.0
        comp = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"3BA %s" % back_comp, "@", "@"]
        expected_ask_calls = [call(f"3TS\r\n"), call(f"3BA?\r\n"), call(f"3TE\r\n"), call(f"3TE\r\n")]
        expected_write_calls = [call(f"3BH{comp}\r\n"), call(f"3PW0\r\n")]
        self.instr.set_hysteresis_compensation(comp, 3)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_hysteresis_compensation_excepts_because_backlash_compensation(self):
        """Test set hysteresis compensation exits because backlash compensation is enabled."""
        state = "32"  # Will call reset to get into CONFIGURATION state
        back_comp = 0.1
        comp = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", f"{self.controller_address}BA %s" % back_comp, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}BA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [
            call(f"{self.controller_address}RS\r\n"),
            call(f"{self.controller_address}PW1\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_hysteresis_compensation(comp)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_hysteresis_compensation_gets_compensation(self):
        """Test get hysteresis compensation."""
        state = "0A"  # in NOT REFERENCED state, no reset needed
        expected_compensation = 10
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}",
            "@",
            f"{self.controller_address}BH %s" % expected_compensation,
            "@",
        ]
        expected_write_calls = [call(f"{self.controller_address}PW1\r\n"), call(f"{self.controller_address}PW0\r\n")]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}BH?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_comp = self.instr.get_hysteresis_compensation()
        self.assertEqual(expected_compensation, actual_comp)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_home_search_type(self):
        """Test getting home search type."""
        state = "14"  # CONFIGURATION state
        expected_type = 2
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}HT %s" % expected_type, "@"]
        expected_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}HT?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_type = self.instr.get_home_search_type()
        self.assertEqual(actual_type, expected_type)

        self._scpi_mock.ask.assert_has_calls(expected_calls)
        self._scpi_mock.write.assert_called_once_with(f"{self.controller_address}PW0\r\n")

    def test_set_home_search_type(self):
        """Test setting home search type."""
        state = "14"  # CONFIGURATION state
        home_type = 1
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]

        self.instr.set_home_search_type(home_type)

        self._scpi_mock.write.assert_any_call(f"1HT{home_type}\r\n")
        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_set_home_search_type_excepts(self):
        """Test setting home search type excepts with type out-of-range."""
        home_type = 5
        self._scpi_mock.ask.return_value = "@"
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_home_search_type(home_type)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_peak_current_limit_gets_limit(self):
        """Test get peak current limit."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        expected_limit = 1.0
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}QIL %s" % expected_limit, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}QIL?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        actual_limit = self.instr.get_peak_current_limit()

        self.assertEqual(expected_limit, actual_limit)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_peak_current_limit_sets_limit(self):
        """Test set peak current limit."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        limit = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        expected_write_calls = [
            call(f"{self.controller_address}QIL%s\r\n" % limit),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_peak_current_limit(limit)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_peak_current_limit_with_error_throws_exception(self):
        """Test set peak current limit with error."""
        out_of_bounds = [0.04999, 3.0001]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_peak_current_limit(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_rms_current_limit_gets_limit(self):
        """Test get rms current limit."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        expected_rms = 10
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}QIR %s" % expected_rms, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}QIR?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        actual_rms = self.instr.get_rms_current_limit()
        self.assertEqual(expected_rms, actual_rms)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_rms_current_limit_sets_limit(self):
        """Test set rms current limit with controller address."""
        state = "0A"  # Set to NOT REFERENCED state
        peak = 2.0
        rms = 0.05
        controller_address = 2
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", f"{controller_address}QIL{peak}", "@", "@"]
        expected_write_calls = [
            call(f"{controller_address}PW1\r\n"),
            call(f"{controller_address}QIR{rms}\r\n"),
            call(f"{controller_address}PW0\r\n"),
        ]
        expected_ask_calls = [
            call(f"{controller_address}TS\r\n"),
            call(f"{controller_address}TE\r\n"),
            call(f"{controller_address}QIL?\r\n"),
            call(f"{controller_address}TE\r\n"),
            call(f"{controller_address}TE\r\n"),
        ]
        self.instr.set_rms_current_limit(rms, controller_address)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_rms_current_limit_sets_limit_under_peak_limit(self):
        """Test set rms current limit works also with lower peak limit."""
        state = "0A"  # Set to NOT REFERENCED state to take the shortest path
        peak = 2.0
        rms = 0.05
        controller_address = 2
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", f"{controller_address}QIL{peak}", "@", "@"]
        expected_write_calls = [call("1PW1\r\n"), call(f"1QIR{rms}\r\n"), call("1PW0\r\n")]
        expected_ask_calls = [call("1TS\r\n"), call("1TE\r\n"), call("1QIL?\r\n"), call("1TE\r\n"), call("1TE\r\n")]
        self.instr.set_rms_current_limit(rms)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_rms_current_limit_excepts_because_peak_limit(self):
        """Test set rms current limit exits because peak limit is lower."""
        state = "0A"  # Set to NOT REFERENCED state
        peak = 1.0
        rms = 1.1
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", f"1QIL{peak}", "@", "@"]
        expected_write_calls = [call("1PW1\r\n"), call("1PW0\r\n")]
        expected_ask_calls = [call("1TS\r\n"), call("1TE\r\n"), call("1QIL?\r\n"), call("1TE\r\n")]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_rms_current_limit(rms)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_rms_current_averaging_time_gets_time(self):
        """Test get rms current averaging time."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        expected_time = 10
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}QIT %s" % expected_time, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}QIT?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        actual_time = self.instr.get_rms_current_averaging_time()

        self.assertEqual(expected_time, actual_time)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_rms_current_averaging_time_sets_time(self):
        """Test set rms current averaging time."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        time = 13
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        expected_write_calls = [
            call(f"{self.controller_address}QIT%i\r\n" % time),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        self._scpi_mock.ask.return_value = "@"
        self.instr.set_rms_current_averaging_time(time)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_rms_current_averaging_time_with_error_throws_exception(self):
        """Test set rms current averaging time with error."""
        out_of_bounds = [0.01, 101]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_rms_current_averaging_time(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_analog_input_value(self):
        """Test getting analog input value."""
        expected_value = 2.0
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}RA %f" % expected_value, "@"]
        expected_calls = [call(f"{self.controller_address}RA\r\n"), call(f"{self.controller_address}TE\r\n")]
        actual_value = self.instr.get_analog_input_value()
        self.assertEqual(actual_value, expected_value)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_get_ttl_input_value(self):
        """Test getting ttl input value."""
        expected_value = 5
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}RB %i" % expected_value, "@"]
        expected_calls = [call(f"{self.controller_address}RB\r\n"), call(f"{self.controller_address}TE\r\n")]
        actual_value = self.instr.get_ttl_input_value()
        self.assertEqual(actual_value, expected_value)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_get_ttl_output_value(self):
        """Test getting ttl output value."""
        expected_value = 3
        self._scpi_mock.ask.side_effect = [f"{self.controller_address}SB %i" % expected_value, "@"]
        expected_calls = [call(f"{self.controller_address}SB?\r\n"), call(f"{self.controller_address}TE\r\n")]
        actual_value = self.instr.get_ttl_output_value()
        self.assertEqual(actual_value, expected_value)

        self._scpi_mock.ask.assert_has_calls(expected_calls)

    def test_set_ttl_output_value(self):
        """Test setting ttl output value."""
        ttl_output = 1
        self._scpi_mock.ask.return_value = "@"
        self.instr.set_ttl_output_value(ttl_output)

        self._scpi_mock.write.assert_called_once_with(f"1SB{ttl_output}\r\n")
        self._scpi_mock.ask.assert_called_once_with("1TE\r\n")

    def test_set_ttl_output_value_excepts(self):
        """Test setting ttl output value excepts when value out of valid range."""
        ttl_output_invalid = [-1, 16]
        for value in ttl_output_invalid:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_ttl_output_value(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_controller_rs485_address(self):
        """Test getting controller rs485 address."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        expected_value = 2
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}SA %s" % expected_value, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SA?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        actual_value = self.instr.get_controller_rs485_address()
        self.assertEqual(actual_value, expected_value)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_controller_rs485_address(self):
        """Test setting ttl output value."""
        state = "14"  # Set to CONFIGURATION state to take the shortest path
        controller_rs485_address = 30
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        expected_write_calls = [
            call(f"{self.controller_address}SA%i\r\n" % controller_rs485_address),
            call(f"{self.controller_address}PW0\r\n"),
        ]
        self.instr.set_controller_rs485_address(controller_rs485_address)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_controller_rs485_address_excepts(self):
        """Test setting ttl output value excepts when value out of valid range."""
        rs485_address_invalid = [1, 32]
        for value in rs485_address_invalid:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_controller_rs485_address(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_negative_software_limit_gets_limit(self):
        """Test get negative software limit."""
        state = "3C"  # Set to DISABLE state to take the shortest path
        expected_limit = -1.0
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}SL %s" % expected_limit, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SL?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_limit = self.instr.get_negative_software_limit()

        self.assertEqual(expected_limit, actual_limit)
        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_negative_software_limit_gets_limit_in_configuration(self):
        """Test get negative software limit in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION
        expected_limit = -1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}",
            f"1TS0000{state}",
            f"{self.controller_address}SL %s" % expected_limit,
            "@",
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SL?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        actual_limit = self.instr.get_negative_software_limit()

        self.assertEqual(expected_limit, actual_limit)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_negative_software_limit_without_persist_sets_limit(self):
        """Test set negative software limit without persist."""
        state = "34"  # Set to READY state
        limit = -1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        expected_write_calls = [
            call(f"{self.controller_address}SL%s\r\n" % limit),
        ]

        self.instr.set_negative_software_limit(limit)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_negative_software_limit_with_persist_sets_limit(self):
        """Test set negative software limit with persist."""
        state = "14"  # Set to CONFIGURATION state
        limit = -1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        expected_write_calls = [
            call(f"{self.controller_address}SL%s\r\n" % limit),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_negative_software_limit(limit, True)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_negative_software_limit_with_error_throws_exception(self):
        """Test set negative software limit with error."""
        out_of_bounds = [0.0001, 1e-12]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_negative_software_limit(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_positive_software_limit_gets_limit(self):
        """Test get positive software limit."""
        state = "3C"  # Set to DISABLE state
        expected_limit = 1.0
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", f"{self.controller_address}SR %s" % expected_limit, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SR?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        actual_limit = self.instr.get_positive_software_limit()

        self.assertEqual(expected_limit, actual_limit)
        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_positive_software_limit_gets_limit_in_configuration(self):
        """Test get positive software limit in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state
        expected_limit = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}",
            f"1TS0000{state}",
            f"{self.controller_address}SR %s" % expected_limit,
            "@",
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SR?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [call(f"{self.controller_address}PW0\r\n")]
        actual_limit = self.instr.get_positive_software_limit()

        self.assertEqual(expected_limit, actual_limit)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_positive_software_limit_without_persist_sets_limit(self):
        """Test set positive software limit without persist."""
        state = "34"  # Set to READY state
        limit = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        expected_write_calls = [
            call(f"{self.controller_address}SR%s\r\n" % limit),
        ]

        self.instr.set_positive_software_limit(limit)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_positive_software_limit_with_persist_sets_limit(self):
        """Test set positive software limit with persist."""
        state = "14"  # Set to CONFIGURATION state
        limit = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [call(f"{self.controller_address}TS\r\n"), call(f"{self.controller_address}TE\r\n")]
        expected_write_calls = [
            call(f"{self.controller_address}SR%s\r\n" % limit),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_positive_software_limit(limit, True)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_positive_software_limit_with_error_throws_exception(self):
        """Test set positive software limit with error."""
        out_of_bounds = [-0.0001, 1e12]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_positive_software_limit(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()



"""A Python wrapper module to wrap around manufacturer provided library header file 'mhdefin.h'. The header file used
is from software version 3.1.0.
Please find the header file in the dedicated software package for the MultiHarp instrument at
https://www.picoquant.com/dl_software/MultiHarp150/MultiHarp150_160_V3_1.zip"""
# This is generated code, do not edit by hand!
import ctypes


_mhlib_function_signatures = [
        ('MH_GetLibraryVersion', ctypes.c_int, [('vers', ctypes.POINTER(ctypes.c_char))]),
        ('MH_GetErrorString', ctypes.c_int, [('errstring', ctypes.POINTER(ctypes.c_char)), ('errcode', ctypes.c_int)]),
        ('MH_OpenDevice', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('MH_CloseDevice', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('MH_Initialize', ctypes.c_int, [('devidx', ctypes.c_int), ('mode', ctypes.c_int), ('refsource', ctypes.c_int)]),
        ('MH_GetHardwareInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('model', ctypes.POINTER(ctypes.c_char)), ('partno', ctypes.POINTER(ctypes.c_char)), ('version', ctypes.POINTER(ctypes.c_char))]),
        ('MH_GetSerialNumber', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('MH_GetFeatures', ctypes.c_int, [('devidx', ctypes.c_int), ('features', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetBaseResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double)), ('binsteps', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetNumOfInputChannels', ctypes.c_int, [('devidx', ctypes.c_int), ('nchannels', ctypes.POINTER(ctypes.c_int))]),
        ('MH_SetSyncDiv', ctypes.c_int, [('devidx', ctypes.c_int), ('div', ctypes.c_int)]),
        ('MH_SetSyncEdgeTrg', ctypes.c_int, [('devidx', ctypes.c_int), ('level', ctypes.c_int), ('edge', ctypes.c_int)]),
        ('MH_SetSyncChannelOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('value', ctypes.c_int)]),
        ('MH_SetSyncDeadTime', ctypes.c_int, [('devidx', ctypes.c_int), ('on', ctypes.c_int), ('deadtime', ctypes.c_int)]),
        ('MH_SetInputEdgeTrg', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('level', ctypes.c_int), ('edge', ctypes.c_int)]),
        ('MH_SetInputChannelOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('value', ctypes.c_int)]),
        ('MH_SetInputDeadTime', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('on', ctypes.c_int), ('deadtime', ctypes.c_int)]),
        ('MH_SetInputChannelEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('enable', ctypes.c_int)]),
        ('MH_SetStopOverflow', ctypes.c_int, [('devidx', ctypes.c_int), ('stop_ovfl', ctypes.c_int), ('stopcount', ctypes.c_uint)]),
        ('MH_SetBinning', ctypes.c_int, [('devidx', ctypes.c_int), ('binning', ctypes.c_int)]),
        ('MH_SetOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('offset', ctypes.c_int)]),
        ('MH_SetHistoLen', ctypes.c_int, [('devidx', ctypes.c_int), ('lencode', ctypes.c_int), ('actuallen', ctypes.POINTER(ctypes.c_int))]),
        ('MH_SetMeasControl', ctypes.c_int, [('devidx', ctypes.c_int), ('control', ctypes.c_int), ('startedge', ctypes.c_int), ('stopedge', ctypes.c_int)]),
        ('MH_SetTriggerOutput', ctypes.c_int, [('devidx', ctypes.c_int), ('period', ctypes.c_int)]),
        ('MH_ClearHistMem', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('MH_StartMeas', ctypes.c_int, [('devidx', ctypes.c_int), ('tacq', ctypes.c_int)]),
        ('MH_StopMeas', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('MH_CTCStatus', ctypes.c_int, [('devidx', ctypes.c_int), ('ctcstatus', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetHistogram', ctypes.c_int, [('devidx', ctypes.c_int), ('chcount', ctypes.POINTER(ctypes.c_uint)), ('channel', ctypes.c_int)]),
        ('MH_GetAllHistograms', ctypes.c_int, [('devidx', ctypes.c_int), ('chcount', ctypes.POINTER(ctypes.c_uint))]),
        ('MH_GetResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double))]),
        ('MH_GetSyncPeriod', ctypes.c_int, [('devidx', ctypes.c_int), ('period', ctypes.POINTER(ctypes.c_double))]),
        ('MH_GetSyncRate', ctypes.c_int, [('devidx', ctypes.c_int), ('syncrate', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetCountRate', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('cntrate', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetAllCountRates', ctypes.c_int, [('devidx', ctypes.c_int), ('syncrate', ctypes.POINTER(ctypes.c_int)), ('cntrates', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetFlags', ctypes.c_int, [('devidx', ctypes.c_int), ('flags', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetElapsedMeasTime', ctypes.c_int, [('devidx', ctypes.c_int), ('elapsed', ctypes.POINTER(ctypes.c_double))]),
        ('MH_GetStartTime', ctypes.c_int, [('devidx', ctypes.c_int), ('timedw2', ctypes.POINTER(ctypes.c_uint)), ('timedw1', ctypes.POINTER(ctypes.c_uint)), ('timedw0', ctypes.POINTER(ctypes.c_uint))]),
        ('MH_GetWarnings', ctypes.c_int, [('devidx', ctypes.c_int), ('warnings', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetWarningsText', ctypes.c_int, [('devidx', ctypes.c_int), ('text', ctypes.POINTER(ctypes.c_char)), ('warnings', ctypes.c_int)]),
        ('MH_SetMarkerHoldoffTime', ctypes.c_int, [('devidx', ctypes.c_int), ('holdofftime', ctypes.c_int)]),
        ('MH_SetMarkerEdges', ctypes.c_int, [('devidx', ctypes.c_int), ('me1', ctypes.c_int), ('me2', ctypes.c_int), ('me3', ctypes.c_int), ('me4', ctypes.c_int)]),
        ('MH_SetMarkerEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('en1', ctypes.c_int), ('en2', ctypes.c_int), ('en3', ctypes.c_int), ('en4', ctypes.c_int)]),
        ('MH_ReadFiFo', ctypes.c_int, [('devidx', ctypes.c_int), ('buffer', ctypes.POINTER(ctypes.c_uint)), ('nactual', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetDebugInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('debuginfo', ctypes.POINTER(ctypes.c_char))]),
        ('MH_GetNumOfModules', ctypes.c_int, [('devidx', ctypes.c_int), ('nummod', ctypes.POINTER(ctypes.c_int))]),
        ('MH_GetModuleInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('modidx', ctypes.c_int), ('modelcode', ctypes.POINTER(ctypes.c_int)), ('versioncode', ctypes.POINTER(ctypes.c_int))]),
        ('MH_WRabbitGetMAC', ctypes.c_int, [('devidx', ctypes.c_int), ('mac_addr', ctypes.POINTER(ctypes.c_char))]),
        ('MH_WRabbitSetMAC', ctypes.c_int, [('devidx', ctypes.c_int), ('mac_addr', ctypes.POINTER(ctypes.c_char))]),
        ('MH_WRabbitGetInitScript', ctypes.c_int, [('devidx', ctypes.c_int), ('initscript', ctypes.POINTER(ctypes.c_char))]),
        ('MH_WRabbitSetInitScript', ctypes.c_int, [('devidx', ctypes.c_int), ('initscript', ctypes.POINTER(ctypes.c_char))]),
        ('MH_WRabbitGetSFPData', ctypes.c_int, [('devidx', ctypes.c_int), ('sfpnames', ctypes.POINTER(ctypes.c_char)), ('dTxs', ctypes.POINTER(ctypes.c_int)), ('dRxs', ctypes.POINTER(ctypes.c_int)), ('alphas', ctypes.POINTER(ctypes.c_int))]),
        ('MH_WRabbitSetSFPData', ctypes.c_int, [('devidx', ctypes.c_int), ('sfpnames', ctypes.POINTER(ctypes.c_char)), ('dTxs', ctypes.POINTER(ctypes.c_int)), ('dRxs', ctypes.POINTER(ctypes.c_int)), ('alphas', ctypes.POINTER(ctypes.c_int))]),
        ('MH_WRabbitInitLink', ctypes.c_int, [('devidx', ctypes.c_int), ('link_on', ctypes.c_int)]),
        ('MH_WRabbitSetMode', ctypes.c_int, [('devidx', ctypes.c_int), ('bootfromscript', ctypes.c_int), ('reinit_with_mode', ctypes.c_int), ('mode', ctypes.c_int)]),
        ('MH_WRabbitSetTime', ctypes.c_int, [('devidx', ctypes.c_int), ('timehidw', ctypes.c_uint), ('timelodw', ctypes.c_uint)]),
        ('MH_WRabbitGetTime', ctypes.c_int, [('devidx', ctypes.c_int), ('timehidw', ctypes.POINTER(ctypes.c_uint)), ('timelodw', ctypes.POINTER(ctypes.c_uint)), ('subsec16ns', ctypes.POINTER(ctypes.c_uint))]),
        ('MH_WRabbitGetStatus', ctypes.c_int, [('devidx', ctypes.c_int), ('wrstatus', ctypes.POINTER(ctypes.c_int))]),
        ('MH_WRabbitGetTermOutput', ctypes.c_int, [('devidx', ctypes.c_int), ('buffer', ctypes.POINTER(ctypes.c_char)), ('nchar', ctypes.POINTER(ctypes.c_int))])
    ]



"""Unit test for Montana Cryostation driver."""

import logging
import math
import unittest
import warnings
from unittest.mock import MagicMock, Mock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException, QMI_InvalidOperationException, QMI_TimeoutException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.montana import Montana_Cryostation


class TestCryostation(unittest.TestCase):

    def setUp(self):
        qmi.start("TestContext")
        self._transport_mock = MagicMock(spec=QMI_TcpTransport)
        with patch(
                'qmi.instruments.montana.cryostation.create_transport',
                return_value=self._transport_mock):
            self.instr: Montana_Cryostation = qmi.make_instrument("instr", Montana_Cryostation, "transport_descriptor")

    def tearDown(self):
        logging.getLogger("qmi.instruments.montana.cryostation").setLevel(logging.NOTSET)
        qmi.stop()

    def _suppress_logging(self):
        logging.getLogger("qmi.instruments.montana.cryostation").setLevel(logging.CRITICAL)

    def test_open_close(self):
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()

    def test_get_alarm_state(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"01", b"F"]

        alarm_state = self.instr.get_alarm_state()
        self.assertFalse(alarm_state)

        self._transport_mock.write.assert_called_once_with(b"03GAS")
        self.assertEqual(self._transport_mock.read.call_args_list, [
            call(2, Montana_Cryostation.RESPONSE_TIMEOUT),
            call(1, Montana_Cryostation.RESPONSE_TIMEOUT)
        ])

        self._transport_mock.read.side_effect = [b"01", b"T"]

        alarm_state = self.instr.get_alarm_state()
        self.assertTrue(alarm_state)

        self.instr.close()

    def test_get_chamber_pressure(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"08", b"660848.6"]
        chamber_pressure = self.instr.get_chamber_pressure("mTorr")
        self.assertAlmostEqual(chamber_pressure, 660848.6)
        self._transport_mock.write.assert_called_once_with(b"03GCP")

        self._transport_mock.read.side_effect = [b"08", b"660848.6"]
        chamber_pressure = self.instr.get_chamber_pressure("Pa")
        self.assertAlmostEqual(chamber_pressure, 660848.6 * (101325.0 / 760000.0))

        self._transport_mock.read.side_effect = [b"08", b"660848.6"]
        chamber_pressure = self.instr.get_chamber_pressure("atm")
        self.assertAlmostEqual(chamber_pressure, 660848.6 / 760000.0)

        self.instr.close()

    def test_get_magnet_state(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"14", b"MAGNET ENABLED"]
        magnet_state = self.instr.get_magnet_state()
        self.assertTrue(magnet_state)
        self._transport_mock.write.assert_called_once_with(b"03GMS")

        self._transport_mock.read.side_effect = [b"15", b"MAGNET DISABLED"]
        magnet_state = self.instr.get_magnet_state()
        self.assertFalse(magnet_state)

        self._transport_mock.read.side_effect = [
            b"83", b"System not able to execute command at this time. Activate the magnet module first. "]
        magnet_state = self.instr.get_magnet_state()
        self.assertIsNone(magnet_state)

        self.instr.close()

    def test_get_magnet_target_field(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"08", b"0.670000"]
        target = self.instr.get_magnet_target_field()
        self.assertAlmostEqual(target, 0.67)
        self._transport_mock.write.assert_called_once_with(b"04GMTF")

        self._transport_mock.read.side_effect = [b"09", b"-9.999999"]
        target = self.instr.get_magnet_target_field()
        self.assertTrue(math.isnan(target))

        self.instr.close()

    def test_get_platform_heater_power(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"05", b"4.904"]
        power = self.instr.get_platform_heater_power()
        self.assertAlmostEqual(power, 4.904)
        self._transport_mock.write.assert_called_once_with(b"04GPHP")

        self._transport_mock.read.side_effect = [b"06", b"-0.100"]
        power = self.instr.get_platform_heater_power()
        self.assertTrue(math.isnan(power))

        self.instr.close()

    def test_get_pid_integral_frequency(self):
        self._transport_mock.read.side_effect = [b"05", b"0.100"]

        self.instr.open()
        v = self.instr.get_pid_integral_frequency()
        self.instr.close()

        self.assertAlmostEqual(v, 0.1)
        self._transport_mock.write.assert_called_once_with(b"05GPIDF")

    def test_get_pid_proportional_gain(self):
        self._transport_mock.read.side_effect = [b"05", b"0.300"]

        self.instr.open()
        v = self.instr.get_pid_proportional_gain()
        self.instr.close()

        self.assertAlmostEqual(v, 0.3)
        self._transport_mock.write.assert_called_once_with(b"05GPIDK")

    def test_get_pid_derivative_time(self):
        self._transport_mock.read.side_effect = [b"05", b"0.100"]

        self.instr.open()
        v = self.instr.get_pid_derivative_time()
        self.instr.close()

        self.assertAlmostEqual(v, 0.1)
        self._transport_mock.write.assert_called_once_with(b"05GPIDT")

    def test_get_platform_stability(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"08", b"10.20239"]
        v = self.instr.get_platform_stability()
        self.assertAlmostEqual(v, 10.20239)
        self._transport_mock.write.assert_called_once_with(b"03GPS")

        self._transport_mock.read.side_effect = [b"08", b"-0.10000"]
        v = self.instr.get_platform_stability()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_get_stage_1_temperature(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"06", b"274.92"]
        v = self.instr.get_stage_1_temperature()
        self.assertAlmostEqual(v, 274.92)
        self._transport_mock.write.assert_called_once_with(b"04GS1T")

        self._transport_mock.read.side_effect = [b"05", b"-0.10"]
        v = self.instr.get_stage_1_temperature()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_get_stage_2_temperature(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"06", b"275.84"]
        v = self.instr.get_stage_2_temperature()
        self.assertAlmostEqual(v, 275.84)
        self._transport_mock.write.assert_called_once_with(b"04GS2T")

        self._transport_mock.read.side_effect = [b"05", b"-0.10"]
        v = self.instr.get_stage_2_temperature()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_get_platform_temperature(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"07", b"289.904"]
        v = self.instr.get_platform_temperature()
        self.assertAlmostEqual(v, 289.904)
        self._transport_mock.write.assert_called_once_with(b"03GPT")

        self._transport_mock.read.side_effect = [b"06", b"-0.100"]
        v = self.instr.get_platform_temperature()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_timeout(self):
        self.instr.open()

        self._transport_mock.read.side_effect = QMI_TimeoutException("too late")
        with self.assertRaises(QMI_TimeoutException):
            v = self.instr.get_platform_temperature()
        self._transport_mock.write.assert_called_once_with(b"03GPT")

        self.instr.close()

    def test_get_sample_stability(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"07", b"0.00900"]
        v = self.instr.get_sample_stability()
        self.assertAlmostEqual(v, 0.009)
        self._transport_mock.write.assert_called_once_with(b"03GSS")

        self._transport_mock.read.side_effect = [b"08", b"-0.10000"]
        v = self.instr.get_sample_stability()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_get_sample_temperature(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"07", b"289.904"]
        v = self.instr.get_sample_temperature()
        self.assertAlmostEqual(v, 289.904)
        self._transport_mock.write.assert_called_once_with(b"03GST")

        self._transport_mock.read.side_effect = [b"06", b"-0.100"]
        v = self.instr.get_sample_temperature()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_get_temperature_setpoint(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"04", b"4.20"]
        v = self.instr.get_temperature_setpoint()
        self.assertAlmostEqual(v, 4.2)
        self._transport_mock.write.assert_called_once_with(b"04GTSP")

        self.instr.close()

    def test_get_user_stability(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"07", b"0.00900"]
        v = self.instr.get_user_stability()
        self.assertAlmostEqual(v, 0.009)
        self._transport_mock.write.assert_called_once_with(b"03GUS")

        self._transport_mock.read.side_effect = [b"08", b"-0.10000"]
        v = self.instr.get_user_stability()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_get_user_temperature(self):
        self.instr.open()

        self._transport_mock.read.side_effect = [b"05", b"3.498"]
        v = self.instr.get_user_temperature()
        self.assertAlmostEqual(v, 3.498)
        self._transport_mock.write.assert_called_once_with(b"03GUT")

        self._transport_mock.read.side_effect = [b"06", b"-0.100"]
        v = self.instr.get_user_temperature()
        self.assertTrue(math.isnan(v))

        self.instr.close()

    def test_set_pid_integral_frequency(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"44", b"OK, Gain = 0.300, Freq = 0.100, Time = 0.100"]
        self.instr.set_pid_integral_frequency(0.1)
        self._transport_mock.write.assert_called_once_with(b"11SPIDF 0.100")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"28", b"Error: Invalid PID parameter"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_pid_integral_frequency(0.1)

        with self.assertRaises(ValueError):
            self.instr.set_pid_integral_frequency(-1)

        with self.assertRaises(ValueError):
            self.instr.set_pid_integral_frequency(101)

        self.instr.close()

    def test_set_pid_proportional_gain(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"44", b"OK, Gain = 0.300, Freq = 0.100, Time = 0.100"]
        self.instr.set_pid_proportional_gain(0.1)
        self._transport_mock.write.assert_called_once_with(b"11SPIDK 0.100")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"28", b"Error: Invalid PID parameter"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_pid_proportional_gain(0.1)

        with self.assertRaises(ValueError):
            self.instr.set_pid_proportional_gain(-1)

        with self.assertRaises(ValueError):
            self.instr.set_pid_proportional_gain(101)

        self.instr.close()

    def test_set_pid_derivative_time(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"44", b"OK, Gain = 0.300, Freq = 0.100, Time = 0.100"]
        self.instr.set_pid_derivative_time(0.1)
        self._transport_mock.write.assert_called_once_with(b"11SPIDT 0.100")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"28", b"Error: Invalid PID parameter"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_pid_derivative_time(0.1)

        with self.assertRaises(ValueError):
            self.instr.set_pid_derivative_time(-1)

        with self.assertRaises(ValueError):
            self.instr.set_pid_derivative_time(101)

        self.instr.close()

    def test_set_temperature_setpoint(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"32", b"OK, Temperature Set Point = 4.20"]
        self.instr.set_temperature_setpoint(5)
        self._transport_mock.write.assert_called_once_with(b"09STSP 5.00")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"24", b"Error: Invalid set point"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_temperature_setpoint(5)

        with self.assertRaises(ValueError):
            self.instr.set_temperature_setpoint(1)

        with self.assertRaises(ValueError):
            self.instr.set_temperature_setpoint(351)

        self.instr.close()

    def test_start_standby(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"02", b"OK"]
        self.instr.start_standby()
        self._transport_mock.write.assert_called_once_with(b"03SSB")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"39", b"System not able to standby at this time"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.start_standby()

        self.instr.close()

    def test_start_warmup(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"02", b"OK"]
        self.instr.start_warmup()
        self._transport_mock.write.assert_called_once_with(b"03SWU")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"39", b"System not able to warmup at this time"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.start_warmup()

        self.instr.close()

    def test_start_cooldown(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"02", b"OK"]
        self.instr.start_cooldown()
        self._transport_mock.write.assert_called_once_with(b"03SCD")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"39", b"System not able to cooldown at this time"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.start_cooldown()

        self.instr.close()

    def test_stop_system(self):
        self._suppress_logging()
        self.instr.open()

        self._transport_mock.read.side_effect = [b"02", b"OK"]
        self.instr.stop_system()
        self._transport_mock.write.assert_called_once_with(b"03STP")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"39", b"System not able to stop at this time"]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.stop_system()

        self.instr.close()

    def test_set_platform_heater_power(self):
        self._suppress_logging()

        self.instr.open()

        # Note: this is probably not the exact string sent by Montana
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]

        self.instr.set_platform_heater_power(0.123)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.1230")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power(1.0)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 1.0000")

        with self.assertRaises(QMI_InvalidOperationException):
            self.instr.set_platform_heater_power(1.1)

        # Turn heater off before closing the instrument.
        # We will separately test scenarios where the instrument is closed with the heater still on.
        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power(0.0)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.0000")

        self.instr.close()

    def test_set_platform_heater_power_failed(self):
        self._suppress_logging()

        self.instr.open()

        # Simulate a failure while trying to set the heater.
        self._transport_mock.read.side_effect = [
            b"18", b"heater out of fuel",
            b"35", b"OK, Platform Heater Power Set Point"
        ]

        # Attempting to set the heater will raise an exception.
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_platform_heater_power(0.123)

        # After the failed attempt to set the heater, the driver will
        # set turn the heater off as a safety measure.
        self.assertEqual(self._transport_mock.write.call_args_list, [
            call(b"11SPHP 0.1230"),
            call(b"11SPHP 0.0000")
        ])

        self.instr.close()

    def test_set_platform_heater_power_burst(self):
        self._suppress_logging()

        self.instr.open()

        # Set the heater to high power.
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(5.0, 2.5)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 5.0000")

        # Then immediately turn the heater back off.
        # We do not test the time limit yet.
        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(0.0, 0.0)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.0000")

        # Test out-of-range values.
        with self.assertRaises(QMI_InvalidOperationException):
            self.instr.set_platform_heater_power_burst(10.1, 1.0)

        with self.assertRaises(QMI_InvalidOperationException):
            self.instr.set_platform_heater_power_burst(1.0, 30.1)

        self.instr.close()

    @patch("qmi.instruments.montana.cryostation.threading.Timer")
    def test_set_platform_heater_power_burst_timelimit(self, timer_mock):
        self._suppress_logging()

        self.instr.open()

        timer_instance_mock = Mock()
        timer_mock.return_value = timer_instance_mock

        # Set the heater to high power.
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(10.0, 1.2)
        self._transport_mock.write.assert_called_once_with(b"12SPHP 10.0000")

        # Driver should have started a background timer thread.
        timer_mock.assert_called_once()
        timer_instance_mock.start.assert_called_once()
        (timer_interval, timer_callback) = timer_mock.call_args[0]
        self.assertEqual(timer_interval, 1.2)

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]

        # Now invoke the timer callback so the driver will think the time limit passed.
        # At this point the driver must turn off the heater.
        timer_callback()
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.0000")

        self.instr.close()

        # The callback thread should be cleaned up when the driver is closed.
        timer_instance_mock.join.assert_called_once()

    @patch("qmi.instruments.montana.cryostation.threading.Timer")
    def test_set_platform_heater_power_burst_extended(self, timer_mock):
        self._suppress_logging()

        self.instr.open()

        timer_instance_mock = Mock()
        timer_mock.return_value = timer_instance_mock

        # Set the heater to high power.
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(5.0, 1.5)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 5.0000")

        # Driver should have started a background timer thread.
        timer_mock.assert_called_once()
        timer_instance_mock.start.assert_called_once()

        timer_instance2_mock = Mock()
        timer_mock.return_value = timer_instance2_mock

        # Set heater power again.
        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(5.0, 1.5)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 5.0000")

        # Driver should have canceled and restarted the timer.
        timer_instance_mock.cancel.assert_called_once()
        timer_instance_mock.join.assert_called_once()
        timer_instance2_mock.start.assert_called_once()

        # Set "safe" heater power with the normal function.
        # This should cancel the timer.
        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power(0.1)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.1000")

        timer_instance2_mock.cancel.assert_called_once()
        timer_instance2_mock.join.assert_called_once()

        # Turn heater off.
        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power(0.0)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.0000")

        self.instr.close()

    def test_set_platform_heater_power_off_on_close(self):
        self.instr.open()

        # Set the heater to high power.
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(5.0, 10.0)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 5.0000")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]

        # Close the driver with the heater still on.
        self.instr.close()

        # The driver should have turned off the heater while closing.
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.0000")

    def test_set_platform_heater_power_off_on_shutdown(self):
        self.instr.open()

        # Set the heater to high power.
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(5.0, 10.0)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 5.0000")

        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]

        # Stop QMI without closing the driver.
        # The driver should turn off the heater during cleanup.
        with warnings.catch_warnings():
            # Suppress warning for stopping QMI without closing instrument.
            warnings.simplefilter("ignore", ResourceWarning)
            qmi.stop()

        # The driver should have turned off the heater during cleanup.
        self._transport_mock.write.assert_called_once_with(b"11SPHP 0.0000")

        # Start QMI again because tearDown() wants to stop it.
        qmi.start("TestContext")

    @patch("time.sleep")
    def test_set_platform_heater_power_off_retry(self, sleep_mock):
        self._suppress_logging()

        self.instr.open()

        # Set the heater to high power.
        self._transport_mock.read.side_effect = [b"35", b"OK, Platform Heater Power Set Point"]
        self.instr.set_platform_heater_power_burst(5.0, 10.0)
        self._transport_mock.write.assert_called_once_with(b"11SPHP 5.0000")

        # Close the driver with the heater still on.
        # The driver will try to turn off the heater, but the first attempt fails.
        self._transport_mock.reset_mock()
        self._transport_mock.read.side_effect = [
            b"34", b"NO, gonna keep heating the diamond",
            b"35", b"OK, Platform Heater Power Set Point"
        ]

        self.instr.close()

        # Expect two attempts to turn off the heater.
        self.assertEqual(self._transport_mock.write.call_args_list, [
            call(b"11SPHP 0.0000"),
            call(b"11SPHP 0.0000")
        ])


if __name__ == '__main__':
    unittest.main()



"""QMI driver for the NKT Photonics Instruments "Boostik" laser amplifier.

The protocol implemented in this driver is documented in the document
'NKT_Photonics_BoostiK_HPA_Users_Manual rev06.pdf', section 4.3.2.
"""

import logging
import struct
from typing import Optional

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class KoherasBoostikLaserAmplifier(QMI_Instrument):
    DEFAULT_TIMEOUT = 0.500  # 500 ms ought to be plenty.

    def __init__(
        self,
        context: QMI_Context,
        name: str,
        transport: str,
        timeout: Optional[float] = DEFAULT_TIMEOUT,
    ):
        super().__init__(context, name)
        self._transport = create_transport(
            transport,
            default_attributes={
                "baudrate": 9600,
                "bytesize": 8,
                "parity": "N",
                "stopbits": 1.0,
                "rtscts": False,
            },
        )

        self._timeout = timeout

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        self._transport.discard_read()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    def _send_command(self, cmd: str) -> None:
        cmd_bin = (cmd + "\r\n").encode("ascii")
        self._transport.write(cmd_bin)

    def _get_string(self, cmd: str) -> str:
        self._send_command(cmd)
        response_bin = self._transport.read_until(b"\r\n", self._timeout)
        response_str = response_bin.decode("ascii")
        assert response_str.endswith("\r\n")
        response_str = response_str[:-2]
        return response_str

    def _get_float(self, cmd: str) -> float:
        response = self._get_string(cmd)
        return float(response)

    def _get_integer(self, cmd: str) -> int:
        response = self._get_string(cmd)
        return int(response)

    def _get_boolean(self, cmd: str) -> bool:
        response = self._get_integer(cmd)
        assert response in [0, 1]
        return bool(response)

    @rpc_method
    def get_current_setpoint(self) -> float:
        """Retrieve the setpoint in ACC (current controlled) mode, in [A]."""
        return self._get_float("ACC")

    @rpc_method
    def get_actual_current(self) -> float:
        """Retrieve the real current value, in [A]."""
        return self._get_float("AMC")

    @rpc_method
    def get_diode_booster_temperature(self) -> float:
        """Retrieve the diode booster temperature, in [°C]."""
        return self._get_float("AMT 1")

    @rpc_method
    def get_ambient_temperature(self) -> float:
        """Retrieve the ambient temperature, in [°C]."""
        return self._get_float("CMA")

    @rpc_method
    def get_input_power(self) -> float:
        """Retrieve input power, in [mW]."""
        return self._get_float("CMP 1")

    @rpc_method
    def get_amplifier_enabled(self) -> bool:
        """Retrieve the ON/OFF status of the amplifier."""
        return self._get_boolean("CDO")

    @rpc_method
    def get_amplifier_information(self) -> str:
        """Retrieve the serial number information of the amplifier."""
        return self._get_string("CDI")

    @rpc_method
    def set_current_setpoint(self, current: float) -> float:
        """Set the ACC current setpoint, in [A].

        The command returns the actual setpoint after executing the command.
        In case the command was unsuccessful (current out of range), the unchanged
        current setpoint is returned.

        Parameters:
            current: In Amperes.
        """
        return self._get_float(f"ACC {current}")

    @rpc_method
    def set_amplifier_enabled(self, enable: bool) -> bool:
        """Enable or disable the amplifier.

        The command returns the enabled status after the command.

        Parameters:
            enable: Boolean for disabling (False) or enabling (True).
        """
        value = int(bool(enable))
        return self._get_boolean(f"CDO {value}")



#! /usr/bin/env python3

import qmi

from qmi.core.pubsub import QMI_SignalReceiver

qmi.start("nsg_client", "qmi.conf")

qmi.show_contexts()
qmi.show_rpc_objects()
qmi.show_instruments()

# The context "nsg_server" provides a simple sample-based interface.
print()
print("Getting some samples from 'nsg_server'...")

nsg = qmi.get_instrument("nsg_server.nsg")
for i in range(10):
    sample = nsg.get_sample()
    print(i, sample)

# The context "nsg_service" provides a pubsub-based interface.
print()
print("Subscribing to 'nsg_service' for a while...")

receiver = QMI_SignalReceiver()
proxy = qmi.get_task("nsg_service.controller")
proxy.sig_sample.subscribe(receiver)
for i in range(10):
    sig = receiver.get_next_signal(timeout=None)
    print(i, "received: {} with arguments t={:.2f} value={:.2f}".format(sig.signal_name, *sig.args))

qmi.stop()



"""
Instrument driver for the Thorlabs TSP01 RevB temperature/humidity sensor.

The Thorlabs TSP01 RevB is not compatible with the original Thorlabs TSP01.
The two instruments have different USB IDs and use different protocols.

This driver supports only the TSP01 RevB (also called TSP01B).
New devices sold by Thorlabs after 2018 are RevB devices.

The TSP01 RevB uses a binary protocol on top of USB-HID.
This driver is based on partial reverse-engineering of the binary protocol.

This driver uses PyUSB and libusb to access the instrument,
and is therefore expected to work only under Linux.
"""

import binascii
import logging
import struct
from typing import List, Optional, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class _Thorlabs_TSP01B_libusb:
    """Exchange binary data packets with the TSP01 RevB sensor via USB HID."""

    # USB vendor ID and product ID of the TSP01 RevB sensor.
    VENDOR_ID = 0x1313
    PRODUCT_ID = 0x80fa

    # Always send/receive packets of 32 bytes.
    PACKET_SIZE = 32

    # Default request timeout (seconds).
    DEFAULT_TIMEOUT = 2.0

    @classmethod
    def list_instruments(cls) -> List[str]:
        """Return serial numbers of attached TSP01 RevB sensors."""

        import usb.core

        instruments = []
        devs = usb.core.find(find_all=True, idVendor=cls.VENDOR_ID, idProduct=cls.PRODUCT_ID)
        for dev in devs:
            try:
                if (dev.manufacturer == "Thorlabs") and (dev.product == "TSP01B"):
                    instruments.append(dev.serial_number)
            except Exception:
                # This happens if the user does not have permission
                # to open the USB device node /dev/usb/<bus>/<addr>.
                _logger.warning("Warning: Can not read USB serial_number")
                continue

        return instruments

    def __init__(self, serial_number: str) -> None:
        """Connect to the instrument.

        Read/write permission to the USB device node is required
        to access the instrument. This is typically accomplished on Linux
        by adding a suitable rule in ``/etc/udev/rules.d``.

        Parameters:
            serial_number: USB serial number of the device to open.

        Raises:
            QMI_InstrumentException: When the specified instrument is not found.
            USBError: If an error occurs at the USB level.
        """

        import usb.core
        import usb.util

        self.timeout = self.DEFAULT_TIMEOUT

        self._dev = usb.core.find(find_all=False,
                                  idVendor=self.VENDOR_ID,
                                  idProduct=self.PRODUCT_ID,
                                  serial_number=serial_number)
        if self._dev is None:
            raise QMI_InstrumentException("Instrument with serial number {} not found (check device permission)"
                                          .format(serial_number))

        # Read current configuration.
        cfg = self._dev.get_active_configuration()

        # Find the HID interface.
        hid_interface = None
        for interface in cfg.interfaces():
            if interface.bInterfaceClass == usb.CLASS_HID:
                hid_interface = interface
                break
        if hid_interface is None:
            raise QMI_InstrumentException("Instrument does not have a HID interface")
        self._hid_interface = hid_interface.index

        # Find the interrupt endpoints.
        endpoint_in = None
        endpoint_out = None
        for endpoint in hid_interface.endpoints():
            if usb.util.endpoint_type(endpoint.bmAttributes) == usb.util.ENDPOINT_TYPE_INTR:
                if ((endpoint_in is None)
                        and (usb.util.endpoint_direction(endpoint.bEndpointAddress) == usb.util.ENDPOINT_IN)):
                    endpoint_in = endpoint.bEndpointAddress
                if ((endpoint_out is None)
                        and (usb.util.endpoint_direction(endpoint.bEndpointAddress) == usb.util.ENDPOINT_OUT)):
                    endpoint_out = endpoint.bEndpointAddress
        if endpoint_in is None:
            raise QMI_InstrumentException("Instrument does not have Interrupt IN endpoint")
        if endpoint_out is None:
            raise QMI_InstrumentException("Instrument does not have Interrupt OUT endpoint")
        self._endpoint_in = endpoint_in
        self._endpoint_out = endpoint_out

        # Detach kernel driver.
        driver_active = self._dev.is_kernel_driver_active(hid_interface.index)
        if driver_active:
            self._dev.detach_kernel_driver(hid_interface.index)

        # Claim the interface.
        usb.util.claim_interface(self._dev, hid_interface.index)

    def close(self) -> None:
        """Release the USB device.

        This object instance must not be used after calling ``close()``.
        """

        import usb.util

        if self._dev is not None:
            usb.util.dispose_resources(self._dev)
        self._dev = None

    def get_feature_report(self) -> bytes:
        """Send USB HID Get Feature Report request and return binary response data."""
        bmRequestType = 0xa1    # interface request
        bRequest = 0x01         # Get Report
        wValue = 0x0300         # Feature Report
        wIndex = self._hid_interface
        wLength = 32
        timeout = int(1000 * self.timeout)
        data = self._dev.ctrl_transfer(bmRequestType,
                                       bRequest,
                                       wValue,
                                       wIndex,
                                       wLength,
                                       timeout)
        return data

    def write_data(self, data: bytes) -> None:
        """Send binary data to the device via USB HID.

        Parameters:
            data: Data to send. This must be a packet of 32 raw bytes.
        """
        timeout = int(1000 * self.timeout)
        self._dev.write(self._endpoint_out, data, timeout)

    def read_data(self) -> bytes:
        """Read binary data from the instrument via USB HID.

        Returns:
            Raw bytes received from the instrument.
        """
        timeout = int(1000 * self.timeout)
        data = self._dev.read(self._endpoint_in, self.PACKET_SIZE, timeout)
        return bytes(data)

    def flush_read(self) -> None:
        """Read any old (pending) replies from the instrument."""

        import usb.core

        # Read until timeout.
        flush_timeout = 0.5
        while True:
            try:
                timeout = int(1000 * flush_timeout)
                _data = self._dev.read(self._endpoint_in, self.PACKET_SIZE, timeout)
            except usb.core.USBError:
                break


class Thorlabs_TSP01B(QMI_Instrument):
    """Instrument driver for the Thorlabs TSP01 RevB temperature/humidity sensor.

    The TSP01 RevB measures temperature and humidity.
    Up to two external temperature probes can be connected to the instrument
    to measure two additional temperature channels.

    The Thorlabs TSP01 RevB is not compatible with the original Thorlabs TSP01.
    This driver supports only the TSP01 RevB (also called TSP01B).
    """

    # Constants for internal use.
    _CHANNEL_TEMPERATURE = 0x00
    _CHANNEL_HUMIDITY = 0x01
    _CHANNEL_EXTERNAL = 0x02

    @staticmethod
    def list_instruments() -> List[str]:
        """Return a list of serial numbers of attached TSP01 RevB instruments."""
        return _Thorlabs_TSP01B_libusb.list_instruments()

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 serial_number: str
                 ) -> None:
        """Initialize the instrument driver.

        Read/write permission to the USB device node is required
        to access the instrument. This is typically accomplished on Linux
        by adding a suitable rule in ``/etc/udev/rules.d``.

        Parameters:
            name: Name for this instrument instance.
            serial_number: USB serial number of the instrument.
        """
        super().__init__(context, name)
        self._serial_number = serial_number
        self._instr: Optional[_Thorlabs_TSP01B_libusb] = None

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        instr = _Thorlabs_TSP01B_libusb(self._serial_number)
        try:
            instr.get_feature_report()
            instr.flush_read()
        except Exception:
            instr.close()
            raise
        self._instr = instr
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        self._check_is_open()
        assert self._instr is not None
        self._instr.close()
        self._instr = None
        super().close()

    @staticmethod
    def _bytes_to_hex(data: bytes) -> str:
        """Convert a byte string to hexadecimal representation.

        This is used to report error messages in case of protocol errors.
        """
        return " ".join(["{:02x}".format(b) for b in data])

    @staticmethod
    def _crc8(data: bytes) -> int:
        """Calculate an 8-bit CRC as used by the TSP01B binary protocol."""
        polynomial = 0x2F
        crc = 0x00
        for b in data:
            crc ^= b
            for i in range(8):
                if crc & 0x80:
                    crc = ((crc & 0x7f) << 1) ^ polynomial
                else:
                    crc = (crc << 1)
        return crc

    @classmethod
    def _format_packet(cls, par1: int, data: bytes) -> bytes:
        """Prepare a binary packet to send to the TSP01B.

        Parameters:
            par1: Command byte.
            data: Optional payload bytes.

        Returns:
            Binary packet of length 32 bytes.
        """

        # Packet format (32 bytes length):
        #   byte 0: fixed byte 0xF0
        #   byte 1: data length
        #   byte 2: fixed byte 0x00
        #   byte 3: fixed byte 0x01
        #   byte 4: command code
        #   byte 5: CRC-8 over first 5 bytes
        #   byte 6: fixed byte 0xF1
        #   starting at byte 7: variable length data (may be empty)
        #   following data: 4 bytes CRC-32 over data (excluding the 0xF1 marker)
        #   padding with 0x00 until total length 32 bytes

        packet_size = 32
        assert 11 + len(data) <= packet_size

        # Prepare header and CRC-8.
        header = bytes([0xF0, len(data), 0x00, 0x01, par1])
        header_crc = cls._crc8(header)

        # Calculate CRC-32 over data.
        data_crc = struct.pack("<I", binascii.crc32(data))

        # Add zero-padding until 32 bytes total length.
        padding = bytes((packet_size - 11 - len(data)) * [0x00])

        # Format final packet.
        packet = header + bytes([header_crc, 0xF1]) + data + data_crc + padding
        return packet

    @classmethod
    def _parse_packet(cls, packet: bytes) -> Tuple[int, bytes]:
        """Decode a binary packet received from the TSP01B.

        Parameters:
            packet: Binary data received from TSP01B.

        Returns:
            Tuple (reply_code, reply_payload).
        """

        # Packet format (32 bytes length):
        #   byte 0: fixed byte 0xF0
        #   byte 1: data length
        #   byte 2: fixed byte 0x00
        #   byte 3: fixed byte 0x01
        #   byte 4: reply code
        #   byte 5: CRC-8 over first 5 bytes
        #   byte 6: fixed byte 0xF1
        #   starting at byte 7: variable length data (may be empty)
        #   following data: 4 bytes CRC-32 over data (excluding the 0xF1 marker)
        #   padding with 0x00 until total length 32 bytes

        # Check minimum packet length.
        if len(packet) < 11:
            raise QMI_InstrumentException(
                "Received short packet from instrument ({})".format(cls._bytes_to_hex(packet)))

        # Check the fixed byte values.
        if (packet[0] != 0xF0) or (packet[6] != 0xF1):
            raise QMI_InstrumentException(
                "Received invalid packet from instrument ({})".format(cls._bytes_to_hex(packet)))

        # Verify the header CRC.
        header_crc = cls._crc8(packet[:5])
        if packet[5] != header_crc:
            raise QMI_InstrumentException(
                "Received bad header CRC from instrument ({})".format(cls._bytes_to_hex(packet)))

        # Check that the data length is consistent with the packet length.
        data_len = packet[1]
        if len(packet) < 11 + data_len:
            raise QMI_InstrumentException(
                "Received invalid packet from instrument ({})".format(cls._bytes_to_hex(packet)))

        # Extract reply code and payload data.
        par1 = packet[4]
        data = packet[7:7+data_len]

        # Verify the data CRC.
        data_crc = struct.pack("<I", binascii.crc32(data))
        if packet[7+data_len:11+data_len] != data_crc:
            raise QMI_InstrumentException(
                "Received bad data CRC from instrument ({})".format(cls._bytes_to_hex(packet)))

        return (par1, data)

    def _query_instrument(self, par1: int, data: bytes) -> Tuple[int, bytes]:
        """Send a query to the instrument and return the response data.

        Parameters:
            par1: Command byte.
            data: Optional payload bytes.

        Returns:
            Tuple (reply_code, reply_payload).
        """
        assert self._instr is not None

        # Send query.
        query_packet = self._format_packet(par1, data)
        self._instr.write_data(query_packet)

        # Receive reply.
        # Typically the instrument sends one dummy reply before sending
        # the actual reply to the command. Dummy replies can be recognized
        # by reply code 0x01 and empty data.
        while True:
            reply = self._instr.read_data()
            (reply_par, reply_data) = self._parse_packet(reply)
            # Skip dummy replies until we receive a non-dummy reply.
            if (reply_par != 0x01) or (len(reply_data) != 0):
                break

        # Send dummy command.
        # The purpose of this dummy command is not known.
        dummy_packet = self._format_packet(0x01, bytes())
        self._instr.write_data(dummy_packet)

        return (reply_par, reply_data)

    def _query_device_info(self, info_type: int) -> str:
        """Query device information.

        Parameters:
            info_type: Device information type code.

        Returns:
            Information string returned by instrument.
        """

        # Send query: command = 0x04, payload data = 1 byte [info_type].
        (reply_par, reply_data) = self._query_instrument(0x04, bytes([info_type]))

        # Expect reply code 0x04, payload data = [info_type] followed by string data.
        if (reply_par != 0x04) or (len(reply_data) < 1) or (reply_data[0] != info_type):
            raise QMI_InstrumentException("Unexpected reply to device info query ({:02x}, {})"
                                          .format(reply_par, self._bytes_to_hex(reply_data)))

        # Find end of string data.
        endpos = reply_data.find(0, 1)
        if endpos <= 0:
            endpos = len(reply_data)
        info_bytes = reply_data[1:endpos]

        # Convert ASCII bytes to string.
        return info_bytes.decode("latin1")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        self._check_is_open()
        device_info_manufacturer_name = 0x03
        device_info_product_name = 0x02
        device_info_serial_number = 0x04
        device_info_firmware_revision = 0x01
        vendor = self._query_device_info(device_info_manufacturer_name)
        model = self._query_device_info(device_info_product_name)
        serial = self._query_device_info(device_info_serial_number)
        version = self._query_device_info(device_info_firmware_revision)
        return QMI_InstrumentIdentification(vendor, model, serial, version)

    @rpc_method
    def reset(self):
        """Reset the instrument."""
        self._check_is_open()
        (reply_par, reply_data) = self._query_instrument(0x04, bytes([0x00]))
        if (reply_par != 0x04) or (reply_data != bytes([0x00])):
            raise QMI_InstrumentException("Unexpected reply to reset command ({:02x}, {})"
                                          .format(reply_par, self._bytes_to_hex(reply_data)))

    def _query_measurement(self, channel: int, ext_channel: int = 0) -> float:
        """Send a measurement query and return the measured value.

        Parameters:
            channel: Channel to measure (one of the `_CHANNEL_xxx` constants).
            ext_channel: Optional external channel selection in case of `_CHANNEL_EXTERNAL` (value 0 or 1).

        Returns:
            Measured value.

        Raises:
            QMI_InstrumentException: If the selected external sensor is not connected.
        """
        assert channel in (self._CHANNEL_TEMPERATURE,
                           self._CHANNEL_HUMIDITY,
                           self._CHANNEL_EXTERNAL)

        # Send query.
        if channel == self._CHANNEL_EXTERNAL:
            query_data = bytes([channel, ext_channel])
        else:
            query_data = bytes([channel])
        (reply_par, reply_data) = self._query_instrument(0x07, query_data)

        # The instrument returns the result as a 32-bit, little-endian floating point number.
        # A specific error code is returned if the external sensor is not connected.
        if (reply_par == 0x13) and (reply_data == bytes([0x03, 0x02, 0x24, 0x00])):
            raise QMI_InstrumentException("Sensor not connected")
        if (reply_par != 0x07) or (len(reply_data) != 5) or (reply_data[0] != channel):
            raise QMI_InstrumentException("Unexpected reply to measurement query ({:02x}, {})"
                                          .format(reply_par, self._bytes_to_hex(reply_data)))
        (value,) = struct.unpack("<f", reply_data[1:5])
        return value

    @rpc_method
    def get_humidity(self) -> float:
        """Perform a humidity measurement.

        Returns:
            Relative humidity as a percentage (range 0 to 100).
        """
        self._check_is_open()
        return self._query_measurement(self._CHANNEL_HUMIDITY)

    @rpc_method
    def get_internal_temperature(self) -> float:
        """Measure the temperature of the internal sensor.

        Returns:
            Temperature in Celsius.
        """
        self._check_is_open()
        return self._query_measurement(self._CHANNEL_TEMPERATURE)

    @rpc_method
    def get_external_temperature(self, ext_channel: int) -> float:
        """Measure the temperature of an external sensor.

        Parameters:
            ext_channel: External sensor channel number (1 or 2).

        Returns:
            Temperature in Celsius.
        """
        if ext_channel not in (1, 2):
            raise ValueError("Unsupported external channel number")
        self._check_is_open()
        return self._query_measurement(self._CHANNEL_EXTERNAL, ext_channel - 1)



"""
NKT Photonics, laser driver and amplifier.

The qmi.instruments.nkt_photonics package provides support for:
- Koheras ADJUSTIK laser driver
- Koheras BOOSTIK laser amplifier.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.nkt_photonics.adjustik import KoherasAdjustikLaser as NktPhotonics_KoherasAdjustik
from qmi.instruments.nkt_photonics.boostik import KoherasBoostikLaserAmplifier as NktPhotonics_KoherasBoostik



from enum import Enum


def update_device_list():
    ...


def get_n_devices():
    ...


def get_device_vendor(d):
    ...


def get_device_model(d):
    ...


def get_device_serial_nbr(d):
    ...


def get_device_address(d):
    ...


def get_device_id(d):
    ...


class BufferPayloadType(Enum):
    CHUNK_DATA = 1
    EXTENDED_CHUNK_DATA = 2


class BufferStatus(Enum):
    SUCCESS = 1
    FAILURE = 2


class AcquisitionMode(Enum):
    CONTINUOUS = 1
    SINGLE_FRAME = 2
    MULTI_FRAME = 3


class PixelFormat(Enum):
    MONO_8 = 0xFFF1
    MONO_16 = 0xFFF2
    MONO_12_PACKED = 0xFFF3
    INVALID = 0


PIXEL_FORMAT_MONO_8 = PixelFormat.MONO_8.value
PIXEL_FORMAT_MONO_16 = PixelFormat.MONO_16.value
PIXEL_FORMAT_MONO_12_PACKED = PixelFormat.MONO_12_PACKED.value
PIXEL_FORMAT_INVALID = PixelFormat.INVALID.value


class Auto:
    def to_string(self, a):
        ...

    def from_string(self, a):
        ...


class GvDevice:
    def get_feature(self, f):
        ...

    def get_device_address(self):
        ...

    def write_register(self, r, v):
        ...


class ValueType:
    name = None


class ChunkParser:
    def get_float_value(self, b, v):
        ...


class GcCommand:
    def execute(self):
        ...

    def get_value_type(self):
        ...


class GcRegisterNode:
    def get_value_as_string(self):
        ...

    def get_value_type(self):
        ...

    def get_value(self):
        ...


class GcEnumeration:
    def get_string_value(self):
        ...

    def set_string_value(self, a):
        ...

    def get_available_string_values(self):
        ...

    def get_value_type(self):
        ...

    def get_value_as_string(self):
        ...


class GcBoolean:
    def get_value(self):
        ...

    def set_value(self, b):
        ...

    def get_value_type(self):
        ...

    def get_value_as_string(self):
        ...


class GcIntegerNode:
    def get_value(self):
        ...

    def set_value(self, b):
        ...

    def get_value_type(self):
        ...

    def get_value_as_string(self):
        ...


class GcFloatNode:
    def get_value_as_string(self):
        ...

    def get_value_type(self):
        ...


class GcRegisterDescriptionNode:
    def get_value_as_string(self):
        ...

    def get_value_type(self):
        ...


class Buffer:
    def new(self, p):
        ...

    def get_image_width(self):
        ...

    def get_image_height(self):
        ...

    def get_image_x(self):
        ...

    def get_image_y(self):
        ...

    def get_image_pixel_format(self):
        ...

    def get_frame_id(self):
        ...

    def get_timestamp(self):
        ...

    def get_data(self):
        ...

    def get_payload_type(self):
        ...

    def get_status(self):
        ...


class Stream:
    def try_pop_buffer(self):
        ...

    def push_buffer(self, b):
        ...

    def connect(self, n, c):
        ...

    def set_emit_signals(self, b):
        ...

    def pop_buffer(self):
        ...

    def timeout_pop_buffer(self, t):
        ...


class Address:
    def get_address(self):
        ...


class InetAddress:
    def to_string(self):
        ...


class Region:
    x = None
    y = None
    width = None
    height = None


class Camera:
    def new(self, d):
        ...

    def start_acquisition(self):
        ...

    def stop_acquisition(self):
        ...

    def get_device(self):
        ...

    def get_vendor_name(self):
        ...

    def get_model_name(self):
        ...

    def get_device_address(self):
        ...

    def get_frame_rate(self):
        ...

    def get_gain(self):
        ...

    def get_gain_auto(self):
        ...

    def get_pixel_format(self):
        ...

    def get_exposure_time(self):
        ...

    def get_exposure_time_auto(self):
        ...

    def get_region(self):
        return Region()

    def get_chunk_mode(self):
        ...

    def get_payload(self):
        ...

    def set_gain(self, g):
        ...

    def set_exposure_time(self, e):
        ...

    def set_pixel_format(self, f):
        ...

    def set_region(self, rx, ry, w, h):
        ...

    def set_acquisition_mode(self, m):
        ...

    def set_frame_count(self, n):
        ...

    def set_frame_rate(self, r):
        ...

    def set_gain_auto(self, g):
        ...

    def set_exposure_time_auto(self, a):
        ...

    def set_chunk_mode(self, c):
        ...

    def create_chunk_parser(self):
        ...

    def create_stream(self, a, b):
        return Stream()
